# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import warnings
from abc import ABC, abstractmethod
from typing import Literal, Optional, Union

import dask.dataframe as dd

from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.gpu_utils import is_cudf_type


class BaseModule(ABC):
    """
    Base class for all NeMo Curator modules.

    Handles validating that data lives on the correct device for each module
    """

    SUPPORTED_BACKENDS = ["pandas", "cudf", "any"]

    def __init__(
        self,
        input_backend: Literal["pandas", "cudf", "any"],
        name: Optional[str] = None,
    ) -> None:
        """
        Constructs a Module

        Args:
            input_backend (Literal["pandas", "cudf", "any"]): The backend the input dataframe must be on for the module to work
            name (str, Optional): The name of the module. If None, defaults to self.__class__.__name__
        """
        super().__init__()
        self.name = name or self.__class__.__name__

        if input_backend not in self.SUPPORTED_BACKENDS:
            raise ValueError(
                f"{input_backend} not one of the supported backends {self.SUPPORTED_BACKENDS}"
            )
        self.input_backend = input_backend

    @abstractmethod
    def call(self, dataset: DocumentDataset):
        """
        Performs an arbitrary operation on a dataset

        Args:
            dataset (DocumentDataset): The dataset to operate on
        """
        raise NotImplementedError("call method must be implemented by subclasses")

    def _validate_correct_backend(self, ddf: dd.DataFrame):
        if self.input_backend == "any":
            return

        backend = "cudf" if is_cudf_type(ddf) else "pandas"
        if backend != self.input_backend:
            raise ValueError(
                f"Module {self.name} requires dataset to have backend {self.input_backend} but got backend {backend}."
                "Try using nemo_curator.ToBackend to swap dataframe backends before running this module."
            )

    def __call__(self, dataset: DocumentDataset):
        """
        Validates the dataset is on the right backend, and performs an arbitrary operation on it

        Args:
            dataset (DocumentDataset): The dataset to operate on
        """
        self._validate_correct_backend(dataset.df)

        return self.call(dataset)


class BaseDeduplicationModule(BaseModule):
    """
    Base class for all NeMo Curator deduplication modules.
    """

    def __init__(
        self,
        id_field: str,
        text_field: str,
        perform_removal: bool = False,
        logger: Union[logging.LoggerAdapter, str] = "./",
        profile_dir: Optional[str] = None,
        cache_dir: Optional[str] = None,
        input_backend: Literal["pandas", "cudf", "any"] = "any",
        **kwargs,
    ):
        super().__init__(input_backend=input_backend, **kwargs)
        self.id_field = id_field
        self.text_field = text_field
        self.perform_removal = perform_removal
        self.logger = logger
        self.profile_dir = profile_dir
        self.cache_dir = cache_dir

        if self.perform_removal and cache_dir is None:
            warnings.warn("cache_dir is recommended to remove duplicates.")

        if cache_dir is None and profile_dir is not None:
            warnings.warn(
                "cache_dir for intermediate outputs is required to generate profiles"
            )

        if not self.perform_removal:
            warnings.warn(
                "In future NeMo Curator releases, the default value for perform_removal will be True."
            )

    @abstractmethod
    def identify_duplicates(self, dataset: DocumentDataset) -> DocumentDataset:
        """
        Identifies duplicates in a dataset

        Args:
            dataset (DocumentDataset): The dataset to identify duplicates in
        """
        raise NotImplementedError(
            "identify_duplicates method must be implemented by subclasses"
        )

    @abstractmethod
    def remove(
        self, dataset: DocumentDataset, duplicates_to_remove: DocumentDataset
    ) -> DocumentDataset:
        """
        Removes duplicates from a dataset

        Args:
            dataset (DocumentDataset): The dataset to remove duplicates from
        """
        raise NotImplementedError("remove method must be implemented by subclasses")

    def call(self, dataset: DocumentDataset) -> DocumentDataset:
        """
        Execute the deduplication process.

        Args:
            dataset (DocumentDataset): Input dataset for deduplication.
        Returns:
            DocumentDataset: Deduplicated dataset if perform_removal is False, otherwise the dataset with duplicates removed.
        """
        duplicates = self.identify_duplicates(dataset)

        if self.perform_removal:
            return self.remove(dataset, duplicates)

        return duplicates



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json
import logging
import shutil

from dask.distributed.utils import silence_logging_cmgr

from nemo_curator.utils.distributed_utils import get_client
from nemo_curator.utils.file_utils import separate_by_metadata
from nemo_curator.utils.script_utils import ArgumentHelper


def main(args):
    print(f"Beginning metadata separation for {args.input_metadata_field}")

    with silence_logging_cmgr(logging.ERROR):
        # Initializes a Dask cluster.
        client = get_client(**ArgumentHelper.parse_client_args(args))

        # Separete corpus by metadata
        metadata_distribution = separate_by_metadata(
            input_data=args.input_data_dir,
            output_dir=args.output_data_dir,
            metadata_field=args.input_metadata_field,
            remove_metadata=args.remove_metadata_field,
            output_type=args.output_file_type,
            input_type=args.input_file_type,
            include_values=args.include_values,
            exclude_values=args.exclude_values,
        )

        # Save metadata distribution to disk
        with open(args.output_metadata_distribution, "w") as fp:
            json.dump(metadata_distribution.compute(), fp)

        # Optionally, remove input directory
        if args.remove_input_dir:
            print(f"Removing all files in {args.input_data_dir}")
            shutil.rmtree(args.input_data_dir)
            print(f"Finished removing all files in {args.input_data_dir}")

        #  Cancel any remaining futures (if any)
        client.cancel(metadata_distribution)

        # Shut down the cluster
        client.shutdown()

        # Close the client
        client.close()


def attach_args(
    parser=argparse.ArgumentParser(
        "Splits a dataset into subdirectories based on metadata values.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
):
    argumentHelper = ArgumentHelper(parser)

    argumentHelper.add_arg_input_data_dir()
    argumentHelper.add_arg_input_file_type()
    argumentHelper.add_arg_output_data_dir(
        help="The output directory to where the metadata-separated files "
        "will be written. Each file will be written to its respective "
        "metadata directory that is a subdirectory of this directory."
    )
    argumentHelper.add_arg_output_file_type()
    argumentHelper.add_distributed_args()
    parser.add_argument(
        "--input-metadata-field",
        type=str,
        default="language",
        help="The name of the field within each datapoint object of the input "
        "file that the dataset should be separated by.",
    )
    parser.add_argument(
        "--output-metadata-distribution",
        type=str,
        help="Output JSON file containing the frequency of documents "
        "that occur for a particular metadata.",
    )
    ArgumentHelper.attach_bool_arg(
        parser,
        "remove-input-dir",
        default=False,
        help="Specify --remove-input-dir to remove the original "
        "input directory. This is false by default.",
    )
    ArgumentHelper.attach_bool_arg(
        parser,
        "remove-metadata-field",
        default=False,
        help="Option of whether to remove the metadata field "
        "after filtering. Useful only in the case in which one metadata "
        "is desired to be separated from the others.",
    )

    exclusive_filters_group = parser.add_mutually_exclusive_group(required=False)
    exclusive_filters_group.add_argument(
        "--include-values",
        nargs="+",
        type=str,
        help="A list of strings representing specific values to be selected or included. "
        "If provided, only the items matching these values should be kept.",
    )
    exclusive_filters_group.add_argument(
        "--exclude-values",
        nargs="+",
        type=str,
        help="A list of strings representing specific values to be excluded or ignored. "
        "If provided, any items matching these values should be skipped.",
    )

    return parser


def console_script():
    main(attach_args().parse_args())



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nemo_curator.modifiers import DocumentModifier

# The system prompt template to be inserted into the documents.
SYS_PROMPT_TEMPLATE = """[INST] <<SYS>> You are reviewing the contents of an email. Based on the content, please categorize this email into one of the following categories:
1. 'Company Business/Strategy.'
2. 'Purely Personal.'
3. 'Personal but in a professional context.'
4. 'Logistic Arrangements.'
5. 'Employment arrangements.'
6. 'Document editing/checking/collaboration.'
Please provide only one category (e.g., 'Purely Personal.'). <</SYS>>

Content::
%s

What should this email be categorized as?
[/INST]
Answer:: """


class AddSystemPrompt(DocumentModifier):
    """
    A simple modifier that adds system prompts to each document.
    """

    def modify_document(self, text: str) -> str:
        """
        Inserts system prompts into the document.

        Args:
            text (str): The text to be modified.

        Returns:
            str: The modified text.
        """
        return SYS_PROMPT_TEMPLATE % text


class AddPeriod(DocumentModifier):
    """
    A simple modifier that adds a period to the end of each email category.
    """

    def modify_document(self, text: str) -> str:
        """
        Adds a period to the end of each email category.

        Args:
            text (str): The text to be modified.

        Returns:
            str: The modified text.
        """
        return text + "."



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

# Disables multiprocessing in torch.compile calls.
# Without this, Dasks multiprocessing combined with PyTorch's
# gives errors like "daemonic processes are not allowed to have children"
# See https://github.com/NVIDIA/NeMo-Curator/issues/31
os.environ["TORCHINDUCTOR_COMPILE_THREADS"] = "1"

from nemo_curator.utils.import_utils import gpu_only_import_from

from .add_id import AddId
from .base import BaseModule
from .config import FuzzyDuplicatesConfig, SemDedupConfig
from .dataset_ops import blend_datasets, Shuffle
from .exact_dedup import ExactDuplicates
from .meta import Sequential
from .modify import Modify
from .splitter import DocumentSplitter
from .joiner import DocumentJoiner
from .task import TaskDecontamination
from .to_backend import ToBackend

# GPU packages
MinHash = gpu_only_import_from("nemo_curator.modules.fuzzy_dedup.minhash", "MinHash")
LSH = gpu_only_import_from("nemo_curator.modules.fuzzy_dedup.lsh", "LSH")
JaccardSimilarity = gpu_only_import_from(
    "nemo_curator.modules.fuzzy_dedup.jaccardsimilarity", "JaccardSimilarity"
)
BucketsToEdges = gpu_only_import_from(
    "nemo_curator.modules.fuzzy_dedup.bucketstoedges", "BucketsToEdges"
)
ConnectedComponents = gpu_only_import_from(
    "nemo_curator.modules.fuzzy_dedup.connectedcomponents", "ConnectedComponents"
)
FuzzyDuplicates = gpu_only_import_from(
    "nemo_curator.modules.fuzzy_dedup.fuzzyduplicates", "FuzzyDuplicates"
)

EmbeddingCreator = gpu_only_import_from(
    "nemo_curator.modules.semantic_dedup.embeddings", "EmbeddingCreator"
)
ClusteringModel = gpu_only_import_from(
    "nemo_curator.modules.semantic_dedup.clusteringmodel", "ClusteringModel"
)
SemanticClusterLevelDedup = gpu_only_import_from(
    "nemo_curator.modules.semantic_dedup.semanticclusterleveldedup",
    "SemanticClusterLevelDedup",
)
SemDedup = gpu_only_import_from(
    "nemo_curator.modules.semantic_dedup.semdedup", "SemDedup"
)

# PyTorch-related imports must come after all imports that require cuGraph
# because of context cleanup issues between PyTorch and cuGraph
# See this issue: https://github.com/rapidsai/cugraph/issues/2718
from .filter import Filter, Score, ScoreFilter, ParallelScoreFilter

__all__ = [
    "AddId",
    "FuzzyDuplicatesConfig",
    "SemDedupConfig",
    "blend_datasets",
    "Shuffle",
    "ExactDuplicates",
    "Filter",
    "Score",
    "ScoreFilter",
    "ParallelScoreFilter",
    "Sequential",
    "Modify",
    "TaskDecontamination",
    "MinHash",
    "LSH",
    "JaccardSimilarity",
    "BucketsToEdges",
    "ConnectedComponents",
    "FuzzyDuplicates",
    "EmbeddingCreator",
    "ClusteringModel",
    "SemanticClusterLevelDedup",
    "SemDedup",
    "BaseModule",
    "ToBackend",
    "DocumentSplitter",
    "DocumentJoiner",
]



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pandas as pd
from dask.dataframe.utils import assert_eq

from nemo_curator import DocumentJoiner
from nemo_curator.datasets import DocumentDataset


class TestDocumentJoiner:
    def test_join_default(self):
        # Input represents documents already split.
        # For example, a document with id=1 split as "a", "b", "c" becomes joined to "a|b|c".
        # Four documents are used.
        data = {
            "id": [1, 1, 1, 2, 3, 3, 4, 4],
            "text": ["a", "b", "c", "nosplit", "start", "middle", "end", ""],
            "segment_id": [0, 1, 2, 0, 0, 1, 0, 1],
        }
        pdf = pd.DataFrame(data)
        dataset = DocumentDataset.from_pandas(pdf, npartitions=1)
        joiner = DocumentJoiner(
            separator="|",
            text_field="text",
            segment_id_field="segment_id",
            document_id_field="id",
            drop_segment_id_field=True,
        )
        result_dataset = joiner(dataset)

        expected_df = pd.DataFrame(
            {"id": [1, 2, 3, 4], "text": ["a|b|c", "nosplit", "start|middle", "end|"]}
        )
        assert_eq(
            result_dataset.df.compute().reset_index(drop=True),
            expected_df,
            check_index=False,
        )

    def test_join_custom_fields(self):
        # Use custom field names:
        #   document id field: "doc"
        #   text field: "content"
        #   segment id field: "s_id"
        # Also keep the segment id field (drop_segment_id_field=False)
        data = {
            "doc": [101, 101, 102, 103, 103, 104, 104],
            "content": ["first", "second", "only", "hello", "world", "baz", ""],
            "s_id": [0, 1, 0, 0, 1, 0, 1],
        }
        pdf = pd.DataFrame(data)
        dataset = DocumentDataset.from_pandas(pdf, npartitions=1)
        joiner = DocumentJoiner(
            separator="~",
            text_field="content",
            segment_id_field="s_id",
            document_id_field="doc",
            drop_segment_id_field=False,
        )
        result_dataset = joiner(dataset)

        # Expected: each document is joined by "~". The segment id becomes the first segment's id.
        expected_df = pd.DataFrame(
            {
                "doc": [101, 102, 103, 104],
                "content": ["first~second", "only", "hello~world", "baz~"],
                "s_id": [0, 0, 0, 0],
            }
        )
        assert_eq(
            result_dataset.df.compute().reset_index(drop=True),
            expected_df,
            check_index=False,
        )

    def test_join_max_length(self):
        # Here we test joining when a maximum length is specified.
        # Each segment carries a precomputed "length" value.
        # The joiner should accumulate segments until adding the next one (plus separator)
        # would exceed max_length=5.
        #
        # For document 1:
        #   segments: "ab"(2), "cd"(2), "ef"(2), "gh"(2)
        #   - "ab" then "cd": 2+2+1 = 5  → join as "ab-cd" (length 5)
        #   - then "ef" then "gh": 2+2+1 = 5 → join as "ef-gh" (length 5)
        #
        # For document 2:
        #   segments: "a"(1), "b"(1) → join as "a-b" (length 3)
        #
        # For document 3:
        #   segment: "hello"(5) → remains "hello"
        #
        # For document 4:
        #   segments: "x"(1), "yz"(2), "0"(1)
        #   - "x" then "yz": 1+2+1 = 4 → "x-yz" (length 4)
        #   - "0" remains alone.
        data = {
            "id": [1, 1, 1, 1, 2, 2, 3, 4, 4, 4],
            "text": ["ab", "cd", "ef", "gh", "a", "b", "hello", "x", "yz", "0"],
            "segment_id": [0, 1, 2, 3, 0, 1, 0, 0, 1, 2],
            "length": [2, 2, 2, 2, 1, 1, 5, 1, 2, 1],
        }
        pdf = pd.DataFrame(data)
        dataset = DocumentDataset.from_pandas(pdf, npartitions=1)
        joiner = DocumentJoiner(
            separator="-",
            text_field="text",
            segment_id_field="segment_id",
            document_id_field="id",
            drop_segment_id_field=True,
            max_length=5,
            length_field="length",
        )
        result_dataset = joiner(dataset)

        expected_df = pd.DataFrame(
            [
                {"id": 1, "text": "ab-cd", "length": 5},
                {"id": 1, "text": "ef-gh", "length": 5},
                {"id": 2, "text": "a-b", "length": 3},
                {"id": 3, "text": "hello", "length": 5},
                {"id": 4, "text": "x-yz", "length": 4},
                {"id": 4, "text": "0", "length": 1},
            ]
        )
        # Sort by id and text to ensure consistent order
        expected_sorted = expected_df.sort_values(by=["id", "text"]).reset_index(
            drop=True
        )
        result_sorted = (
            result_dataset.df.compute()
            .sort_values(by=["id", "text"])
            .reset_index(drop=True)
        )
        assert_eq(result_sorted, expected_sorted, check_index=False)

    def test_join_with_string_ids(self):
        # Test join functionality when document id field is a string.
        data = {
            "doc": ["doc1", "doc1", "doc2", "doc3", "doc3", "doc4", "doc4"],
            "text": ["a", "b", "nosplit", "start", "middle", "end", ""],
            "segment_id": [0, 1, 0, 0, 1, 0, 1],
        }
        pdf = pd.DataFrame(data)
        dataset = DocumentDataset.from_pandas(pdf, npartitions=1)
        joiner = DocumentJoiner(
            separator="|",
            text_field="text",
            segment_id_field="segment_id",
            document_id_field="doc",
            drop_segment_id_field=True,
        )
        result_dataset = joiner(dataset)

        expected_df = pd.DataFrame(
            {
                "doc": ["doc1", "doc2", "doc3", "doc4"],
                "text": ["a|b", "nosplit", "start|middle", "end|"],
            }
        )
        assert_eq(
            result_dataset.df.compute().reset_index(drop=True),
            expected_df,
            check_index=False,
        )



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import time

import cudf
import dask_cudf
import numpy as np

from nemo_curator import LSH
from nemo_curator.datasets import DocumentDataset
from nemo_curator.log import create_logger
from nemo_curator.utils.distributed_utils import get_client
from nemo_curator.utils.fuzzy_dedup_utils.id_mapping import convert_str_id_to_int
from nemo_curator.utils.script_utils import ArgumentHelper


def pre_imports():
    import cudf  # noqa: F401


def main(args):
    logger = create_logger(
        rank=0, log_file=os.path.join(args.log_dir, "rank_000.log"), name="lsh_log"
    )
    logger.info(f"Starting workflow with args:\n {args}")

    assert args.device == "gpu"
    client = get_client(**ArgumentHelper.parse_client_args(args))
    logger.info(f"Client Created {client}")
    client.run(pre_imports)
    logger.info("Pre imports complete")

    data_paths = args.input_data_dirs
    id_field = args.input_json_id_field
    minhash_field = args.input_minhash_field

    dfs = []
    for data_path in data_paths:
        dfs.append(
            dask_cudf.read_parquet(data_path, blocksize="2GB", aggregate_files=True)
        )
    df = dask_cudf.concat(dfs, ignore_unknown_divisions=True)
    df = df[~df[id_field].isna()]
    df = df.map_partitions(
        convert_str_id_to_int,
        id_column=id_field,
        meta=cudf.DataFrame(
            {minhash_field: [[1, 2, 3]], "doc_id": [1], "dataset_id": np.uint32(1)}
        ),
    )

    lsh = LSH(
        cache_dir=args.output_bucket_dir,
        num_hashes=args.minhash_length,
        num_buckets=args.num_bands,
        buckets_per_shuffle=args.buckets_per_shuffle,
        id_fields=["dataset_id", "doc_id"],
        profile_dir=args.profile_path,
        minhash_field=minhash_field,
        false_positive_check=args.false_positive_check,
        logger=logger,
    )

    t1 = time.time()
    _ = lsh(DocumentDataset(df))
    logger.info(f"Computing and writing buckets took {time.time() - t1} s")


def attach_args():
    description = """
Computes buckets from existing minhashes and writes the output
to files. Each row corresponds to a document ID, followed by the columns
denoting the bucket IDs to which the document belongs.
    """
    parser = argparse.ArgumentParser(
        description,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    argumentHelper = ArgumentHelper(parser)

    argumentHelper.parse_gpu_dedup_args()
    argumentHelper.add_arg_minhash_length()
    parser.add_argument(
        "--buckets-per-shuffle",
        type=int,
        required=True,
        help="Number of buckets to shuffle per batch.",
    )
    parser.add_argument(
        "--input-minhash-field",
        type=str,
        default="_minhash_signature",
        help="Name of the column containing minhashes.",
    )
    parser.add_argument(
        "--num-bands",
        type=int,
        default=20,
        help="The number of minhashes to compute for each document.",
    )
    parser.add_argument(
        "--output-bucket-dir",
        type=str,
        required=True,
        help="Output directory where minhashes will be written. "
        "Each Parquet file consists of document and bucket IDs.",
    )
    parser.add_argument(
        "--false-positive-check",
        action="store_true",
        help="Converts LSH buckets to integers required for running the false positive check",
    )

    return parser


def console_script():
    main(attach_args().parse_args())


if __name__ == "__main__":
    main(attach_args().parse_args())



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pandas as pd
from dask.dataframe.utils import assert_eq

from nemo_curator import DocumentJoiner, DocumentSplitter
from nemo_curator.datasets import DocumentDataset


class TestSplitJoinReconstruction:
    def test_reconstruction_default(self):
        # Create an original dataset with a unique "id" column and text examples.
        # Four examples include edge cases:
        #   "a|b|c"          → multiple splits
        #   "nosplit"        → no separator present
        #   "a||b|"          → consecutive separators yield empty strings
        #   ""               → empty document
        docs = ["a|b|c", "nosplit", "a||b|", ""]
        pdf = pd.DataFrame({"id": [1, 2, 3, 4], "text": docs})
        original_dataset = DocumentDataset.from_pandas(pdf, npartitions=1)

        # First, split using "|" as separator.
        splitter = DocumentSplitter(separator="|")
        split_dataset = splitter(original_dataset)

        # Then, rejoin using the same separator.
        joiner = DocumentJoiner(
            separator="|",
            text_field="text",
            segment_id_field="segment_id",
            document_id_field="id",
            drop_segment_id_field=True,
        )
        reconstructed_dataset = joiner(split_dataset)

        # The reconstructed "text" column should match the original.
        original_sorted = (
            original_dataset.df.compute().sort_values(by="id").reset_index(drop=True)
        )
        reconstructed_sorted = (
            reconstructed_dataset.df.compute()
            .sort_values(by="id")
            .reset_index(drop=True)
        )
        assert_eq(reconstructed_sorted, original_sorted, check_index=False)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nemo_curator.modifiers import DocumentModifier


class QuotationRemover(DocumentModifier):
    """
    Removes quotations from a document following a few rules:
    - If the document is less than 2 characters, it is returned unchanged.
    - If the document starts and ends with a quotation mark and there are
        no newlines in the document, the quotation marks are removed.
    - If the document starts and ends with a quotation mark and there are
        newlines in the document, the quotation marks are removed only if
        the first line does not end with a quotation mark.
    """

    def __init__(self):
        super().__init__()

    def modify_document(self, text: str) -> str:
        if len(text.strip()) > 2 and text[0] == '"' and text[-1] == '"':
            if "\n" not in text.strip():
                text = text[1:-1]
            elif text.split("\n")[0][-1] != '"':
                text = text[1:-1]
        return text



import json
import os
import time

import requests
from openai import OpenAI


class Embed:
    def __init__(self):
        self.client = OpenAI(
            base_url="https://ai.api.nvidia.com/v1/retrieval/nvidia",
            api_key=os.getenv("BUILD_NVIDIA_API_KEY"),
        )

    def invoke(self, text):
        return (
            self.client.embeddings.create(
                input=[text],
                model="NV-Embed-QA",
                encoding_format="float",
                extra_body={"input_type": "query", "truncate": "NONE"},
            )
            .data[0]
            .embedding
        )


class LLaMa_405B:
    def __init__(self):
        self.url = "https://integrate.api.nvidia.com/v1/chat/completions"
        self.headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer " + os.getenv("BUILD_NVIDIA_API_KEY"),
        }

    def invoke(self, prompt, schema=None):
        self.payload = {
            "model": "meta/llama-3.1-405b-instruct",
            "messages": [{"role": "user", "content": prompt}],
            "temperature": 0,
            "stream": False,
            "max_tokens": 1024,
        }
        if schema is not None:
            self.payload["nvext"] = schema

        session = requests.Session()
        response = session.post(self.url, headers=self.headers, json=self.payload)

        while response.status_code == 202:
            request_id = response.headers.get("NVCF-REQID")
            fetch_url = fetch_url_format + request_id
            response = session.get(fetch_url, headers=headers)

        response_body = response.json()

        return response_body["choices"][0]["message"]["content"]



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import os
from abc import ABC, abstractmethod
from typing import List, Literal, Optional, Tuple, Union

import dask.dataframe as dd
import pandas as pd
from dask import compute, delayed

from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.distributed_utils import read_single_partition


class DocumentDownloader(ABC):
    """Abstract class for downloading remote data to disk"""

    def __init__(self):
        super().__init__()

    @abstractmethod
    def download(self, url):
        pass


class DocumentIterator(ABC):
    """
    Abstract iterator class for reading in raw records that have been
    downloaded to disk
    """

    def __init__(self):
        super().__init__()

    @abstractmethod
    def iterate(self, file_path):
        pass


class DocumentExtractor(ABC):
    """Abstract class for extracting text from records read from disk"""

    def __init__(self):
        super().__init__()

    @abstractmethod
    def extract(self, content):
        pass


def import_downloader(downloader_path):
    module_path, downloader_name = downloader_path.rsplit(".", 1)
    downloader_module = importlib.import_module(module_path)
    downloader_class = getattr(downloader_module, downloader_name)
    if not issubclass(downloader_class, DocumentDownloader):
        raise ValueError(
            f"Input downloader {downloader_class.__name__} "
            "must be derived from DocumentDownloader defined in "
            "nemo_curator.download.docbuilder"
        )
    return downloader_class


def import_iterator(iterator_path):
    module_path, iterator_name = iterator_path.rsplit(".", 1)
    iterator_module = importlib.import_module(module_path)
    iterator_class = getattr(iterator_module, iterator_name)
    if not issubclass(iterator_class, DocumentIterator):
        raise ValueError(
            f"Input iterator {iterator_class.__name__} "
            "must be derived from DocumentIterator "
            "defined in nemo_curator.download.docbuilder"
        )
    return iterator_class


def import_extractor(extractor_path):
    module_path, extractor_name = extractor_path.rsplit(".", 1)
    extractor_module = importlib.import_module(module_path)
    extractor_class = getattr(extractor_module, extractor_name)
    if not issubclass(extractor_class, DocumentExtractor):
        raise ValueError(
            f"Input extractor {extractor_class.__name__} "
            "must be derived from DocumentExtractor defined "
            "in nemo_curator.download.docbuilder"
        )
    return extractor_class


def _download_and_extract_single_partition(
    paths: List[Tuple[str, str]],
    downloader: DocumentDownloader,
    iterator: DocumentIterator,
    extractor: DocumentExtractor,
    output_type: Literal["jsonl", "parquet"],
    keep_raw_download: bool,
    force_download: bool,
    input_meta: Union[str, dict] = None,
    filename_col: str = "file_name",
    record_limit: Optional[int] = None,
) -> pd.DataFrame:
    """
    Downloads a single partition from a URL and extracts its contents in-memory without writing
    the extracted output to disk. The provided output_path is used only to check if an extracted
    output already exists and, if so, skips re-downloading and extraction.

    Parameters:
        paths (Tuple[str, str]): A tuple (url, output_path) where 'url' is the source URL and
            'output_path' is the expected location of a previously extracted output.
        downloader (DocumentDownloader): An object to download the content from the URL.
        iterator (DocumentIterator): An object to iterate over records in the downloaded file.
        extractor (DocumentExtractor): An object to extract the desired content from each record.
        output_type (Literal["jsonl", "parquet"]): The output file format/extension. Must be either "jsonl" or "parquet". Defaults to "jsonl". This parameter is only used to verify whether an extracted output already exists.
        keep_raw_download (bool): If False, deletes the raw download file after extraction.
        force_download (bool): If False and output_path exists, skips downloading and extraction.
        input_meta (Union[str, dict], optional): Metadata describing the input file's structure.
        filename_col (str, optional): Name of the column to store the filename within the result DataFrame.
        record_limit (int, optional): Limit the number of records to extract from each file.
    Returns:
        pd.DataFrame: A DataFrame containing the extracted records.
    """
    url, output_path = paths

    # If an extracted output already exists and we're not forcing a download, load and return it.
    if os.path.exists(output_path) and not force_download:
        partition = read_single_partition(
            [output_path],
            backend="pandas",
            file_type=output_type,
            add_filename=filename_col,
        )
        return partition

    # Download the file and extract its records in memory.
    downloaded_file = downloader.download(url)
    records = []
    for item in iterator.iterate(downloaded_file):
        if record_limit is not None and len(records) >= record_limit:
            break
        record_meta, content = item
        extracted = extractor.extract(content)
        if extracted is not None:
            # Merge the extracted data and record metadata into one dictionary.
            line = {**extracted, **record_meta}
            records.append(line)

    partition = pd.DataFrame(records)
    # Add a filename column for consistency using the basename of the output_path.
    partition[filename_col] = os.path.basename(output_path)

    # Since we're not writing the extracted partition to disk, the output_path is not used here.
    # Clean up the raw downloaded file if it's not meant to be kept.
    if not keep_raw_download and os.path.exists(downloaded_file):
        os.remove(downloaded_file)

    return partition


def download_and_extract(
    urls: List[str],
    output_paths: List[str],
    downloader: DocumentDownloader,
    iterator: DocumentIterator,
    extractor: DocumentExtractor,
    output_format: dict,
    output_type: Literal["jsonl", "parquet"] = "jsonl",
    keep_raw_download: bool = False,
    force_download: bool = False,
    input_meta: Union[str, dict] = None,
    filename_col: str = "file_name",
    record_limit: Optional[int] = None,
) -> DocumentDataset:
    """
    Download files from the given URLs, extract their records, and
    construct a DocumentDataset.

    For each URL provided, this function downloads the corresponding
    file (unless an extracted output already exists and force_download is
    False), iterates over its records, extracts the desired content, and
    finally converts all records into a DocumentDataset.

    Args:
        urls (List[str]):
            A list of URLs from which to download dataset files.
        output_paths (List[str]):
            A list of file paths where the extracted outputs should be
            found. If a file already exists at a given path and force_download
            is False, that partition is skipped.
        downloader (DocumentDownloader):
            The downloader instance responsible for fetching files from
            the specified URLs.
        iterator (DocumentIterator):
            The iterator instance used to traverse the downloaded file
            and yield records.
        extractor (DocumentExtractor):
            The extractor instance used to obtain the desired content from
            each record.
        output_format (dict):
            A dictionary mapping column names to the data types for the
            extracted records.
        output_type (Literal["jsonl", "parquet"], optional):
            The output file format/extension. Must be either "jsonl" or "parquet".
            Defaults to "jsonl". This parameter is only used to verify whether
            an extracted output already exists.
        keep_raw_download (bool, optional):
            If True, the raw downloaded files are retained after extraction.
            Defaults to False.
        force_download (bool, optional):
            If False and an output file already exists at a given path, the
            download and extraction for that file are skipped.
            Defaults to False.
        input_meta (Union[str, dict], optional):
            Optional metadata describing the input file's schema.
            Defaults to None.
        filename_col (str, optional):
            The name for the column in the resulting dataset that records
            the basename of the output file. Defaults to "file_name".
        record_limit (int, optional): Limit the number of records to extract from each file.
            Defaults to None.
    Returns:
        DocumentDataset:
            A dataset composed of the records extracted from the downloaded
            files.
    """
    # Validate parameters
    if not urls:
        raise ValueError("No URLs were provided to download")
    if len(urls) != len(output_paths):
        raise ValueError(
            f"Different number of URLs and output_paths. {len(urls)} URLs vs {len(output_paths)} output_paths"
        )

    # Ensure consistent ordering of output_format keys.
    output_format = dict(sorted(output_format.items()))

    df = dd.from_map(
        _download_and_extract_single_partition,
        zip(urls, output_paths),
        downloader=downloader,
        iterator=iterator,
        extractor=extractor,
        output_type=output_type,
        keep_raw_download=keep_raw_download,
        force_download=force_download,
        enforce_metadata=False,
        input_meta=input_meta,
        filename_col=filename_col,
        record_limit=record_limit,
        meta=output_format,
    )

    return DocumentDataset(df)


def batch_download(urls: List[str], downloader: DocumentDownloader) -> List[str]:
    """
    Downloads all the urls using the downloader in parallel
    """
    delayed_downloads = [delayed(downloader.download)(url) for url in urls]

    return compute(*delayed_downloads)



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

from nemo_curator.utils.distributed_utils import get_client
from nemo_curator.utils.file_utils import reshard_jsonl
from nemo_curator.utils.script_utils import ArgumentHelper


def main(args):
    client = get_client(**ArgumentHelper.parse_client_args(args))

    reshard_jsonl(
        args.input_data_dir,
        args.output_resharded_dir,
        output_file_size=args.output_file_size,
        start_index=args.start_index,
        file_prefix=args.prefix,
    )


def attach_args(
    parser=argparse.ArgumentParser(
        """
Makes balanced text files of output size --block-size from
a directory of input files. The output files will be renamed
as output_dir/000.jsonl, output_dir/001.jsonl, etc. Users
may specify the desired number of output files and the block size
will be computed from this specified quantity.

The size of the input files must be larger than the specified
--block-size.
""",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
):
    argumentHelper = ArgumentHelper(parser)

    argumentHelper.add_arg_input_data_dir()
    argumentHelper.add_distributed_args()
    parser.add_argument(
        "--output-file-size",
        type=str,
        default="100M",
        help="Approximate size of output files. Must specify with a string and "
        "with the unit K, M or G for kilo, mega, or gigabytes.",
    )
    parser.add_argument(
        "--output-resharded-dir",
        type=str,
        default=None,
        required=True,
        help="Output directory to where the sharded .jsonl files will be written.",
    )
    parser.add_argument(
        "--prefix",
        type=str,
        default="",
        help="Prefix to use to prepend to output file number.",
    )
    parser.add_argument(
        "--start-index",
        type=int,
        default=0,
        help="Starting index for naming the output files.",
    )

    return parser


def console_script():
    main(attach_args().parse_args())



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re
from pathlib import Path

import pandas as pd
from dask import dataframe as dd
from dask.distributed import Client, LocalCluster

import nemo_curator as nc
from nemo_curator.datasets import DocumentDataset
from nemo_curator.filters import DocumentFilter
from nemo_curator.modifiers import PiiModifier
from nemo_curator.pii.algorithm import PiiDeidentifier
from nemo_curator.utils.decorators import batched

LOGGER = logging.getLogger(__name__)


def load_test_cases(filename):
    filepath = Path("tests/pii_data") / filename
    with open(filepath) as fp:
        data = fp.readlines()

    raw_data = [
        (re.sub(r"<[^>]*>([^<]*)</[^>]*>", r"\1", line)).strip() for line in data
    ]
    masked_data = [
        (
            re.sub(r"(<[^>]*>([^<]*)</[^>]*>)", lambda x: "*" * len(x.group(2)), line)
        ).strip()
        for line in data
    ]

    return list(zip(raw_data, masked_data))


def compare_outputs(output1, output2):
    output1 = re.sub(r"\*[\* ]*", "*****", output1)
    output2 = re.sub(r"\*[\* ]*", "*****", output2)
    return output1 == output2


def generate_single_category_test(category, filename):
    deidentifier = PiiDeidentifier("en", [category], "mask")
    test_data = load_test_cases(filename)

    for input, target in test_data:
        output = deidentifier.deidentify_text(input)
        print("============================")
        print("Input: ", input)
        print("Output: ", output)
        print("Expected Output: ", target)
        print("Matches:", "No" if output != target else "Yes")
        assert output == target


class TestPIIAccuracy:
    def test_email(self):
        generate_single_category_test("EMAIL_ADDRESS", "emails.txt")

    def test_ip_address(self):
        generate_single_category_test("IP_ADDRESS", "ip_address.txt")

    def test_address(self):
        generate_single_category_test("ADDRESS", "address.txt")

    def test_ssn(self):
        generate_single_category_test("US_SSN", "ssn.txt")

    def test_birthdates(self):
        generate_single_category_test("DATE_TIME", "birthdates.txt")

    def test_card_no(self):
        generate_single_category_test("CREDIT_CARD", "card_no.txt")

    def test_names(self):
        generate_single_category_test("PERSON", "names.txt")

    def test_phone_numbers(self):
        generate_single_category_test("PHONE_NUMBER", "phone_numbers.txt")

    def test_multiple(self):
        deidentifier = PiiDeidentifier("en", anonymize_action="mask")
        test_data = load_test_cases("multiple.txt")

        for input, target in test_data:
            output = deidentifier.deidentify_text(input)
            print("============================")
            print("Input: ", input)
            print("Output: ", output)
            print("Expected Output: ", target)
            match = compare_outputs(output, target)
            output1 = re.sub(r"\*[\* ]*", "*****", output)
            output2 = re.sub(r"\*[\* ]*", "*****", target)
            print(output1)
            print(output2)
            print("match value: ", match)
            print("Matches:", "No" if not match else "Yes")
            assert match == True

    def test_batch_accuracy(self):
        deidentifier = PiiDeidentifier("en", anonymize_action="mask")
        test_data = load_test_cases("multiple.txt")
        inputs = [data[0] for data in test_data]
        targets = [data[1] for data in test_data]
        outputs = deidentifier.deidentify_text_batch(inputs)
        print("Inputs: ", inputs)
        print("Output: ", outputs)
        print("Expected Outputs: ", targets)

        match = all(compare_outputs(x, y) for x, y in zip(outputs, targets))
        print("Matches:", "No" if not match else "Yes")
        assert match == True


class BatchedLengthFilter(DocumentFilter):
    """
    Keeps documents of a given length
    """

    def __init__(self, min_length=5, max_length=10):
        super().__init__()
        self.min_length = min_length
        self.max_length = max_length

    @batched
    def score_document(self, df):
        return df.str.len()

    @batched
    def keep_document(self, scores):
        min_threshold = self.min_length <= scores
        max_threshold = scores <= self.max_length
        return min_threshold & max_threshold


class TestPIIModule:
    def test_filter_chain(self):
        inputs = [
            "Alice goes on a walk",
            "Bob goes on a walk",
            "Someone named Charlie goes on a walk",
            "A human walking is David",
            "A human walking is Eliza",
        ]
        targets = [
            "***** goes on a walk",
            "*** goes on a walk",
            "A human walking is *****",
            "A human walking is *****",
        ]
        input_df = pd.DataFrame({"text": inputs})
        target_df = pd.DataFrame({"text": targets})
        with LocalCluster(n_workers=1, threads_per_worker=1) as cluster:
            with Client(cluster):
                input_dataset = DocumentDataset(dd.from_pandas(input_df, npartitions=1))
                pipeline = nc.Sequential(
                    [
                        nc.ScoreFilter(
                            BatchedLengthFilter(min_length=0, max_length=25)
                        ),
                        nc.Modify(
                            PiiModifier(
                                language="en", anonymize_action="mask", device="cpu"
                            )
                        ),
                    ]
                )
                output_dataset = pipeline(input_dataset)

                output_df = output_dataset.df.compute().reset_index(drop=True)
                match = all(output_df["text"] == target_df["text"])
                assert match



import tempfile

import pandas as pd
import pytest

from nemo_curator._compat import DASK_CUDF_PARQUET_READ_INCONSISTENT_SCHEMA
from nemo_curator.utils.distributed_utils import (
    read_data,
    read_data_blocksize,
    read_data_files_per_partition,
)

NUM_FILES = 5
NUM_RECORDS = 100
import os
from typing import Dict, List, Literal, Optional

import numpy as np


# Fixture to create multiple small JSONL files
@pytest.fixture
def mock_multiple_jsonl_files(tmp_path):
    file_paths = []
    for file_id in range(NUM_FILES):
        jsonl_file = tmp_path / f"test_{file_id}.jsonl"
        with open(jsonl_file, "w") as f:
            for record_id in range(NUM_RECORDS):
                # 100 rows are ~5kb
                f.write(
                    f'{{"id": "id_{file_id}_{record_id}", "text": "A longish string {file_id}_{record_id}"}}\n'
                )
        file_paths.append(str(jsonl_file))
    return file_paths


# Fixture to create multiple small Parquet files
@pytest.fixture
def mock_multiple_parquet_files(tmp_path):
    file_paths = []
    for file_id in range(NUM_FILES):
        # 100 rows are ~5kb
        parquet_file = tmp_path / f"test_{file_id}.parquet"
        df = pd.DataFrame(
            [
                {
                    "id": f"id_{file_id}_{record_id}",
                    "text": f"A string {file_id}_{record_id}",
                }
                for record_id in range(NUM_RECORDS)
            ]
        )
        # We specify row_group_size so that we can test splitting a single big file into smaller chunks
        df.to_parquet(parquet_file, compression=None, row_group_size=10)
        file_paths.append(str(parquet_file))
    return file_paths


# Fixture to create arbitrary npy files to test custom read functions
@pytest.fixture
def mock_npy_files(tmp_path):
    file_paths = []
    for file_id in range(NUM_FILES):
        npy_file = tmp_path / f"test_{file_id}.npy"
        np.save(npy_file, np.asarray([file_id], dtype=np.float32))
        file_paths.append(str(npy_file))
    return file_paths


@pytest.fixture
def mock_multiple_jsonl_files_different_cols(tmp_path):
    file_paths = []
    for file_id in range(NUM_FILES):
        jsonl_file = tmp_path / f"different_cols_test_{file_id}.jsonl"

        def make_record_without_meta(file_id, record_id):
            return {
                "id": f"id_{file_id}_{record_id}",
                "text": f"A string {file_id}_{record_id}",
            }

        def make_record_with_meta(file_id, record_id):
            return {
                "text": f"A string {file_id}_{record_id}",
                "meta1": [
                    {"field1": "field_one", "field2": "field_two"},
                ],
                "id": f"id_{file_id}_{record_id}",
            }

        df = pd.DataFrame(
            [
                (
                    make_record_without_meta(file_id, record_id)
                    if file_id == 0
                    else make_record_with_meta(file_id, record_id)
                )
                for record_id in range(NUM_RECORDS)
            ]
        )

        df.to_json(jsonl_file, orient="records", lines=True)
        file_paths.append(str(jsonl_file))
    return file_paths


# Fixture to create multiple small Parquet files
@pytest.fixture
def mock_multiple_parquet_files_different_cols(tmp_path):
    file_paths = []
    for file_id in range(NUM_FILES):
        # 100 rows are ~5kb
        parquet_file = tmp_path / f"test_diff_cols_{file_id}.parquet"

        def make_record_without_meta(file_id, record_id):
            return {
                "id": f"id_{file_id}_{record_id}",
                "text": f"A string {file_id}_{record_id}",
            }

        def make_record_with_meta(file_id, record_id):
            return {
                "text": f"A string {file_id}_{record_id}",
                "meta1": [
                    {"field1": "field_one", "field2": "field_two"},
                ],
                "id": f"id_{file_id}_{record_id}",
            }

        df = pd.DataFrame(
            [
                (
                    make_record_without_meta(file_id, record_id)
                    if file_id == 0
                    else make_record_with_meta(file_id, record_id)
                )
                for record_id in range(NUM_RECORDS)
            ]
        )
        df.to_parquet(parquet_file, compression=None, row_group_size=10)
        file_paths.append(str(parquet_file))
    return file_paths


@pytest.mark.gpu
@pytest.mark.parametrize("file_type", ["jsonl", "parquet"])
@pytest.mark.parametrize("blocksize", ["1kb", "5kb", "10kb"])
def test_cudf_read_data_blocksize_partitioning(
    mock_multiple_jsonl_files, mock_multiple_parquet_files, file_type, blocksize
):
    import cudf

    input_files = (
        mock_multiple_jsonl_files
        if file_type == "jsonl"
        else mock_multiple_parquet_files
    )

    df = read_data_blocksize(
        input_files=input_files,
        backend="cudf",
        file_type=file_type,
        blocksize=blocksize,
        add_filename=False,
        input_meta=None,
        columns=None,
    )

    # Compute the number of partitions in the resulting DataFrame
    num_partitions = df.optimize().npartitions
    # Assert that we have two partitions (since we have ~15KB total data and a blocksize of 10KB)
    if blocksize == "1kb":
        assert (
            num_partitions > NUM_FILES
        ), f"Expected > {NUM_FILES} partitions but got {num_partitions}"
    elif blocksize == "5kb":
        assert (
            num_partitions == NUM_FILES
        ), f"Expected {NUM_FILES} partitions but got {num_partitions}"
    elif blocksize == "10kb":
        assert (
            num_partitions < NUM_FILES
        ), f"Expected < {NUM_FILES} partitions but got {num_partitions}"
    else:
        raise ValueError(f"Invalid blocksize: {blocksize}")
    total_rows = len(df)
    assert (
        total_rows == NUM_FILES * NUM_RECORDS
    ), f"Expected {NUM_FILES * NUM_RECORDS} rows but got {total_rows}"

    assert isinstance(df["id"].compute(), cudf.Series)


@pytest.mark.parametrize("file_type", ["jsonl", "parquet"])
@pytest.mark.parametrize("blocksize", ["1kb", "5kb", "10kb"])
def test_pandas_read_data_blocksize_partitioning(
    mock_multiple_jsonl_files, mock_multiple_parquet_files, file_type, blocksize
):
    input_files = (
        mock_multiple_jsonl_files
        if file_type == "jsonl"
        else mock_multiple_parquet_files
    )

    df = read_data_blocksize(
        input_files=input_files,
        backend="pandas",
        file_type=file_type,
        blocksize=blocksize,
        add_filename=False,
        input_meta=None,
        columns=None,
    )

    # Compute the number of partitions in the resulting DataFrame
    num_partitions = df.npartitions
    # Our total data is ~25kb where each file is 5kb
    if blocksize == "1kb":
        assert (
            num_partitions > NUM_FILES
        ), f"Expected > {NUM_FILES} partitions but got {num_partitions}"
    elif blocksize == "5kb":
        assert (
            num_partitions == NUM_FILES
        ), f"Expected {NUM_FILES} partitions but got {num_partitions}"
    elif blocksize == "10kb":
        # Because pandas doesn't suppport reading json files together, a partition will only be as big as a single file
        if file_type == "jsonl":
            assert (
                num_partitions == NUM_FILES
            ), f"Expected {NUM_FILES} partitions but got {num_partitions}"
        # Parquet files can be read together
        elif file_type == "parquet":
            assert (
                num_partitions < NUM_FILES
            ), f"Expected > {NUM_FILES} partitions but got {num_partitions}"
    else:
        raise ValueError(f"Invalid blocksize: {blocksize}")
    total_rows = len(df)
    assert (
        total_rows == NUM_FILES * NUM_RECORDS
    ), f"Expected {NUM_FILES * NUM_RECORDS} rows but got {total_rows}"

    assert isinstance(df["id"].compute(), pd.Series)


@pytest.mark.parametrize(
    "backend",
    ["pandas", pytest.param("cudf", marks=pytest.mark.gpu)],
)
@pytest.mark.parametrize("file_type", ["jsonl", "parquet"])
@pytest.mark.parametrize("fpp", [1, NUM_FILES // 2, NUM_FILES, NUM_FILES * 2])
def test_read_data_fpp_partitioning(
    mock_multiple_jsonl_files, mock_multiple_parquet_files, backend, file_type, fpp
):
    input_files = (
        mock_multiple_jsonl_files
        if file_type == "jsonl"
        else mock_multiple_parquet_files
    )

    df = read_data_files_per_partition(
        input_files=input_files,
        backend=backend,
        file_type=file_type,
        files_per_partition=fpp,
        add_filename=False,
        input_meta=None,
        columns=None,
    )

    # Compute the number of partitions in the resulting DataFrame
    num_partitions = df.npartitions
    # Assert that we have two partitions (since we have ~15KB total data and a blocksize of 10KB)
    if fpp == 1:
        assert (
            num_partitions == NUM_FILES
        ), f"Expected {NUM_FILES} partitions but got {num_partitions}"
    elif fpp == NUM_FILES // 2:
        assert (
            num_partitions < NUM_FILES
        ), f"Expected {NUM_FILES} partitions but got {num_partitions}"
    elif fpp >= NUM_FILES:
        assert num_partitions == 1, f"Expected 1 partition but got {num_partitions}"
    else:
        raise ValueError(f"Invalid fpp: {fpp}")
    total_rows = len(df)
    assert (
        total_rows == NUM_FILES * NUM_RECORDS
    ), f"Expected {NUM_FILES * NUM_RECORDS} rows but got {total_rows}"
    if backend == "cudf":
        import cudf

        assert isinstance(df["id"].compute(), cudf.Series)
    elif backend == "pandas":
        assert isinstance(df["id"].compute(), pd.Series)


@pytest.mark.parametrize(
    "backend",
    [
        "pandas",
        pytest.param("cudf", marks=pytest.mark.gpu),
    ],
)
@pytest.mark.parametrize("filename_arg", [True, "some_filename"])
def test_read_data_blocksize_add_filename_jsonl(
    mock_multiple_jsonl_files, backend, filename_arg
):
    df = read_data_blocksize(
        input_files=mock_multiple_jsonl_files,
        backend=backend,
        file_type="jsonl",
        blocksize="128Mib",
        add_filename=filename_arg,
        input_meta=None,
        columns=None,
    )

    filename_str = "file_name" if filename_arg is True else filename_arg
    assert filename_str in df.columns
    file_names = df[filename_str].unique().compute()
    if backend == "cudf":
        file_names = file_names.to_pandas()

    assert len(file_names) == NUM_FILES
    assert set(file_names.values) == {
        f"test_{file_id}.jsonl" for file_id in range(NUM_FILES)
    }


@pytest.mark.parametrize(
    "backend",
    [
        "pandas",
        pytest.param("cudf", marks=pytest.mark.gpu),
    ],
)
@pytest.mark.parametrize("filename_arg", [True, "some_filename"])
def test_read_data_blocksize_add_filename_parquet(
    mock_multiple_parquet_files, backend, filename_arg
):
    with pytest.raises(
        ValueError,
        match="add_filename and blocksize cannot be set at the same time for Parquet files",
    ):
        read_data_blocksize(
            input_files=mock_multiple_parquet_files,
            backend=backend,
            file_type="parquet",
            blocksize="128Mib",
            add_filename=filename_arg,
            input_meta=None,
            columns=None,
        )


@pytest.mark.parametrize(
    "backend,file_type",
    [
        pytest.param("cudf", "jsonl", marks=pytest.mark.gpu),
        pytest.param("cudf", "parquet", marks=pytest.mark.gpu),
        ("pandas", "jsonl"),
        ("pandas", "parquet"),
    ],
)
@pytest.mark.parametrize("filename_arg", [True, "some_filename"])
def test_read_data_fpp_add_filename(
    mock_multiple_jsonl_files,
    mock_multiple_parquet_files,
    backend,
    file_type,
    filename_arg,
):
    input_files = (
        mock_multiple_jsonl_files
        if file_type == "jsonl"
        else mock_multiple_parquet_files
    )

    df = read_data_files_per_partition(
        input_files=input_files,
        backend=backend,
        file_type=file_type,
        files_per_partition=NUM_FILES,
        add_filename=filename_arg,
        input_meta=None,
        columns=None,
    )

    filename_str = "file_name" if filename_arg is True else filename_arg
    assert filename_str in df.columns
    assert list(df.columns) == list(df.head().columns)
    assert set(df.columns) == {filename_str, "id", "text"}
    file_names = df[filename_str].unique().compute()
    if backend == "cudf":
        file_names = file_names.to_pandas()

    assert len(file_names) == NUM_FILES
    assert set(file_names.values) == {
        f"test_{file_id}.{file_type}" for file_id in range(NUM_FILES)
    }


@pytest.mark.parametrize(
    "backend",
    [
        "pandas",
        pytest.param("cudf", marks=pytest.mark.gpu),
    ],
)
@pytest.mark.parametrize(
    "file_type,add_filename,function_name",
    [
        *[
            ("jsonl", True, func)
            for func in ["read_data_blocksize", "read_data_files_per_partition"]
        ],
        *[
            ("jsonl", False, func)
            for func in ["read_data_blocksize", "read_data_files_per_partition"]
        ],
        *[
            ("parquet", False, func)
            for func in ["read_data_blocksize", "read_data_files_per_partition"]
        ],
        *[("parquet", True, "read_data_files_per_partition")],
    ],
)
@pytest.mark.parametrize(
    "cols_to_select", [None, ["id"], ["text", "id"], ["id", "text"]]
)
def test_read_data_select_columns(
    mock_multiple_jsonl_files,
    mock_multiple_parquet_files,
    backend,
    file_type,
    add_filename,
    function_name,
    cols_to_select,
):
    input_files = (
        mock_multiple_jsonl_files
        if file_type == "jsonl"
        else mock_multiple_parquet_files
    )
    if function_name == "read_data_files_per_partition":
        func = read_data_files_per_partition
        read_kwargs = {"files_per_partition": 1}
    elif function_name == "read_data_blocksize":
        func = read_data_blocksize
        read_kwargs = {"blocksize": "128Mib"}

    df = func(
        input_files=input_files,
        backend=backend,
        file_type=file_type,
        add_filename=add_filename,
        input_meta=None,
        columns=list(cols_to_select) if cols_to_select else None,
        **read_kwargs,
    )
    if not cols_to_select:
        cols_to_select = ["id", "text"]

    assert list(df.columns) == list(df.head().columns)
    if not add_filename:
        assert list(df.columns) == sorted(cols_to_select)
    else:
        assert list(df.columns) == sorted(cols_to_select + ["file_name"])


@pytest.mark.parametrize(
    "backend",
    [
        "pandas",
        pytest.param("cudf", marks=pytest.mark.gpu),
    ],
)
@pytest.mark.parametrize(
    "function_name", ["read_data_blocksize", "read_data_files_per_partition"]
)
@pytest.mark.parametrize(
    "input_meta", [{"id": "str"}, {"text": "str"}, {"id": "str", "text": "str"}]
)
def test_read_data_input_meta(
    mock_multiple_jsonl_files, backend, function_name, input_meta
):
    if function_name == "read_data_files_per_partition":
        func = read_data_files_per_partition
        read_kwargs = {"files_per_partition": 1}
    elif function_name == "read_data_blocksize":
        func = read_data_blocksize
        read_kwargs = {"blocksize": "128Mib"}

    df = func(
        input_files=mock_multiple_jsonl_files,
        backend=backend,
        file_type="jsonl",
        add_filename=False,
        input_meta=input_meta,
        columns=None,
        **read_kwargs,
    )

    assert list(df.columns) == list(input_meta.keys())


""" Tests below this test for custom read functions """


@pytest.mark.parametrize(
    "backend", ["pandas", pytest.param("cudf", marks=pytest.mark.gpu)]
)
def test_read_data_custom_read_function(mock_npy_files, backend):
    # This function ignores file_type, add_filename, columns, and input_meta
    def read_npy_file(files: List[str], backend: Literal["cudf", "pandas"], **kwargs):
        if backend == "cudf":
            import cudf as df_backend
            import cupy as arr_backend
        else:
            import numpy as arr_backend
            import pandas as df_backend

        df = df_backend.DataFrame(
            [(os.path.basename(file), arr_backend.load(file)) for file in files],
            columns=["id", "embedding"],
        )
        return df

    expected_df = pd.DataFrame(
        [
            {"id": f"test_{file_id}.npy", "embedding": np.asarray([file_id])}
            for file_id in range(NUM_FILES)
        ],
    )

    # Test that we can read the file without specifying columns
    df_no_columns = read_data(
        input_files=mock_npy_files,
        backend=backend,
        file_type="npy",
        read_func_single_partition=read_npy_file,
    )
    assert df_no_columns.optimize().npartitions == NUM_FILES
    df_no_columns_computed = df_no_columns.to_backend("pandas").compute()
    pd.testing.assert_frame_equal(
        df_no_columns_computed.sort_values("id").reset_index(drop=True),
        expected_df[["embedding", "id"]],  # because we sort columns by name
    )

    # Test multiple files per partition
    df_fpp_2 = read_data(
        input_files=mock_npy_files,
        backend=backend,
        file_type="npy",
        read_func_single_partition=read_npy_file,
        files_per_partition=2,
    )
    assert df_fpp_2.optimize().npartitions == int(np.ceil(NUM_FILES / 2))
    df_fpp_2_computed = df_fpp_2.to_backend("pandas").compute()
    pd.testing.assert_frame_equal(
        df_fpp_2_computed.sort_values("id").reset_index(drop=True),
        expected_df[["embedding", "id"]],  # because we sort columns by name,
    )


""" Tests below this test for inconsistent schema """


def xfail_inconsistent_schema_jsonl():
    return pytest.mark.xfail(
        reason="inconsistent schemas are not supported with jsonl files, "
        "see https://github.com/dask/dask/issues/11595"
    )


@pytest.mark.parametrize(
    "backend",
    [
        pytest.param("pandas"),
        pytest.param("cudf", marks=[pytest.mark.gpu]),
    ],
)
@pytest.mark.parametrize("file_type", ["jsonl", "parquet"])
@pytest.mark.parametrize("fpp", [1, 3, 5])
def test_read_data_different_columns_files_per_partition(
    mock_multiple_jsonl_files_different_cols,
    mock_multiple_parquet_files_different_cols,
    backend,
    file_type,
    fpp,
):
    read_kwargs = {"columns": ["id", "text"]}
    if file_type == "jsonl":
        input_files = mock_multiple_jsonl_files_different_cols
        read_kwargs["input_meta"] = {"id": "str", "text": "str"}
    elif file_type == "parquet":
        input_files = mock_multiple_parquet_files_different_cols
        if backend == "cudf":
            read_kwargs["allow_mismatched_pq_schemas"] = True

    df = read_data(
        input_files=input_files,
        file_type=file_type,
        backend=backend,
        add_filename=False,
        files_per_partition=fpp,
        blocksize=None,
        **read_kwargs,
    )
    assert list(df.columns) == ["id", "text"]
    assert list(df.compute().columns) == ["id", "text"]
    with tempfile.TemporaryDirectory() as tmpdir:
        df.to_parquet(tmpdir)
    assert len(df) == NUM_FILES * NUM_RECORDS


@pytest.mark.skip(reason="Parquet tests are failing after upgrading to RAPIDS 25.02")
@pytest.mark.parametrize(
    "backend,file_type",
    [
        pytest.param(
            "cudf", "jsonl", marks=[pytest.mark.gpu, xfail_inconsistent_schema_jsonl()]
        ),
        pytest.param("pandas", "jsonl", marks=[xfail_inconsistent_schema_jsonl()]),
        pytest.param(
            "cudf",
            "parquet",
            marks=[pytest.mark.gpu]
            + (
                [xfail_inconsistent_schema_jsonl()]
                if not DASK_CUDF_PARQUET_READ_INCONSISTENT_SCHEMA
                else []
            ),
        ),
        pytest.param("pandas", "parquet"),
    ],
)
@pytest.mark.parametrize("blocksize", ["1kb", "5kb", "10kb"])
def test_read_data_different_columns_blocksize(
    mock_multiple_jsonl_files_different_cols,
    mock_multiple_parquet_files_different_cols,
    backend,
    file_type,
    blocksize,
):
    read_kwargs = {"columns": ["id", "text"]}
    read_kwargs["columns"] = ["id", "text"]
    if file_type == "jsonl":
        input_files = mock_multiple_jsonl_files_different_cols
        read_kwargs["input_meta"] = {"id": "str", "text": "str"}
    elif file_type == "parquet":
        input_files = mock_multiple_parquet_files_different_cols
        if backend == "cudf":
            read_kwargs["allow_mismatched_pq_schemas"] = True

    df = read_data(
        input_files=input_files,
        file_type=file_type,
        blocksize=blocksize,
        files_per_partition=None,
        backend=backend,
        add_filename=False,
        **read_kwargs,
    )
    assert list(df.columns) == ["id", "text"]
    assert list(df.compute().columns) == ["id", "text"]
    with tempfile.TemporaryDirectory() as tmpdir:
        df.to_parquet(tmpdir)
    assert len(df) == NUM_FILES * NUM_RECORDS



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This file is adapted from cuML's safe_imports module:
# https://github.com/rapidsai/cuml/blob/e93166ea0dddfa8ef2f68c6335012af4420bc8ac/python/cuml/internals/safe_imports.py


import importlib
import logging
import traceback
from contextlib import contextmanager

from nemo_curator.utils.gpu_utils import GPU_INSTALL_STRING

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())


class UnavailableError(Exception):
    """Error thrown if a symbol is unavailable due to an issue importing it"""


@contextmanager
def null_decorator(*args, **kwargs):
    if len(kwargs) == 0 and len(args) == 1 and callable(args[0]):
        return args[0]
    else:

        def inner(func):
            return func

        return inner


class UnavailableMeta(type):
    """A metaclass for generating placeholder objects for unavailable symbols

    This metaclass allows errors to be deferred from import time to the time
    that a symbol is actually used in order to streamline the usage of optional
    dependencies. This is particularly useful for attempted imports of GPU-only
    modules which will only be invoked if GPU-only functionality is
    specifically used.

    If an attempt to import a symbol fails, this metaclass is used to generate
    a class which stands in for that symbol. Any attempt to call the symbol
    (instantiate the class) or access its attributes will throw an
    UnavailableError exception. Furthermore, this class can be used in
    e.g. isinstance checks, since it will (correctly) fail to match any
    instance it is compared against.

    In addition to calls and attribute access, a number of dunder methods are
    implemented so that other common usages of imported symbols (e.g.
    arithmetic) throw an UnavailableError, but this is not guaranteed for
    all possible uses. In such cases, other exception types (typically
    TypeErrors) will be thrown instead.
    """

    def __new__(meta, name, bases, dct):
        if dct.get("_msg", None) is None:
            dct["_msg"] = f"{name} could not be imported"
        name = f"MISSING{name}"
        return super(UnavailableMeta, meta).__new__(meta, name, bases, dct)

    def __call__(cls, *args, **kwargs):
        raise UnavailableError(cls._msg)

    def __getattr__(cls, name):
        # Special case for unittest.mock
        # When unittest.mock is checking if an object is async, it checks for __func__
        # We need to return None for this specific attribute to allow mocking
        if name == "__func__":
            return None
        # Also handle other attributes that unittest.mock might check
        if name in ("__await__", "__aenter__", "__aexit__", "__aiter__", "__anext__"):
            return None
        raise UnavailableError(cls._msg)

    def __eq__(cls, other):
        raise UnavailableError(cls._msg)

    def __lt__(cls, other):
        raise UnavailableError(cls._msg)

    def __gt__(cls, other):
        raise UnavailableError(cls._msg)

    def __ne__(cls, other):
        raise UnavailableError(cls._msg)

    def __abs__(cls, other):
        raise UnavailableError(cls._msg)

    def __add__(cls, other):
        raise UnavailableError(cls._msg)

    def __radd__(cls, other):
        raise UnavailableError(cls._msg)

    def __iadd__(cls, other):
        raise UnavailableError(cls._msg)

    def __floordiv__(cls, other):
        raise UnavailableError(cls._msg)

    def __rfloordiv__(cls, other):
        raise UnavailableError(cls._msg)

    def __ifloordiv__(cls, other):
        raise UnavailableError(cls._msg)

    def __lshift__(cls, other):
        raise UnavailableError(cls._msg)

    def __rlshift__(cls, other):
        raise UnavailableError(cls._msg)

    def __mul__(cls, other):
        raise UnavailableError(cls._msg)

    def __rmul__(cls, other):
        raise UnavailableError(cls._msg)

    def __imul__(cls, other):
        raise UnavailableError(cls._msg)

    def __ilshift__(cls, other):
        raise UnavailableError(cls._msg)

    def __pow__(cls, other):
        raise UnavailableError(cls._msg)

    def __rpow__(cls, other):
        raise UnavailableError(cls._msg)

    def __ipow__(cls, other):
        raise UnavailableError(cls._msg)

    def __rshift__(cls, other):
        raise UnavailableError(cls._msg)

    def __rrshift__(cls, other):
        raise UnavailableError(cls._msg)

    def __irshift__(cls, other):
        raise UnavailableError(cls._msg)

    def __sub__(cls, other):
        raise UnavailableError(cls._msg)

    def __rsub__(cls, other):
        raise UnavailableError(cls._msg)

    def __isub__(cls, other):
        raise UnavailableError(cls._msg)

    def __truediv__(cls, other):
        raise UnavailableError(cls._msg)

    def __rtruediv__(cls, other):
        raise UnavailableError(cls._msg)

    def __itruediv__(cls, other):
        raise UnavailableError(cls._msg)

    def __divmod__(cls, other):
        raise UnavailableError(cls._msg)

    def __rdivmod__(cls, other):
        raise UnavailableError(cls._msg)

    def __neg__(cls):
        raise UnavailableError(cls._msg)

    def __invert__(cls):
        raise UnavailableError(cls._msg)

    def __hash__(cls):
        raise UnavailableError(cls._msg)

    def __index__(cls):
        raise UnavailableError(cls._msg)

    def __iter__(cls):
        raise UnavailableError(cls._msg)

    def __delitem__(cls, name):
        raise UnavailableError(cls._msg)

    def __setitem__(cls, name, value):
        raise UnavailableError(cls._msg)

    def __enter__(cls, *args, **kwargs):
        raise UnavailableError(cls._msg)

    def __get__(cls, *args, **kwargs):
        raise UnavailableError(cls._msg)

    def __delete__(cls, *args, **kwargs):
        raise UnavailableError(cls._msg)

    def __len__(cls):
        raise UnavailableError(cls._msg)


def is_unavailable(obj):
    """Helper to check if given symbol is actually a placeholder"""
    return type(obj) is UnavailableMeta


class UnavailableNullContext:
    """A placeholder class for unavailable context managers

    This context manager will return a value which will throw an
    UnavailableError if used in any way, but the context manager itself can be
    safely invoked.
    """

    def __init__(self, *args, **kwargs):
        pass

    def __enter__(self):
        return UnavailableMeta(
            "MissingContextValue",
            (),
            {"_msg": "Attempted to make use of placeholder context return value."},
        )

    def __exit__(self, *args, **kwargs):
        pass


def safe_import(module, *, msg=None, alt=None):
    """A function used to import modules that may not be available

    This function will attempt to import a module with the given name, but it
    will not throw an ImportError if the module is not found. Instead, it will
    return a placeholder object which will raise an exception only if used.

    Parameters
    ----------
    module: str
        The name of the module to import.
    msg: str or None
        An optional error message to be displayed if this module is used
        after a failed import.
    alt: object
        An optional module to be used in place of the given module if it
        fails to import

    Returns
    -------
    object
        The imported module, the given alternate, or a class derived from
        UnavailableMeta.
    """
    try:
        return importlib.import_module(module)
    except ImportError:
        exception_text = traceback.format_exc()
        logger.debug(f"Import of {module} failed with: {exception_text}")
    except Exception:
        exception_text = traceback.format_exc()
        raise
    if msg is None:
        msg = f"{module} could not be imported"
    if alt is None:
        return UnavailableMeta(module.rsplit(".")[-1], (), {"_msg": msg})
    else:
        return alt


def safe_import_from(module, symbol, *, msg=None, alt=None):
    """A function used to import symbols from modules that may not be available

    This function will attempt to import a symbol with the given name from
    the given module, but it will not throw an ImportError if the symbol is not
    found. Instead, it will return a placeholder object which will raise an
    exception only if used.

    Parameters
    ----------
    module: str
        The name of the module in which the symbol is defined.
    symbol: str
        The name of the symbol to import.
    msg: str or None
        An optional error message to be displayed if this symbol is used
        after a failed import.
    alt: object
        An optional object to be used in place of the given symbol if it fails
        to import

    Returns
    -------
    object
        The imported symbol, the given alternate, or a class derived from
        UnavailableMeta.
    """
    try:
        imported_module = importlib.import_module(module)
        return getattr(imported_module, symbol)
    except ImportError:
        exception_text = traceback.format_exc()
        logger.debug(f"Import of {module} failed with: {exception_text}")
    except AttributeError:
        exception_text = traceback.format_exc()
        logger.info(f"Import of {symbol} from {module} failed with: {exception_text}")
    except Exception:
        exception_text = traceback.format_exc()
        raise
    if msg is None:
        msg = f"{module}.{symbol} could not be imported"
    if alt is None:
        return UnavailableMeta(symbol, (), {"_msg": msg})
    else:
        return alt


def gpu_only_import(module, *, alt=None):
    """A function used to import modules required only in GPU installs

    This function will attempt to import a module with the given name.
    This function will attempt to import a symbol with the given name from
    the given module, but it will not throw an ImportError if the symbol is not
    found. Instead, it will return a placeholder object which will raise an
    exception only if used with instructions on installing a GPU build.

    Parameters
    ----------
    module: str
        The name of the module to import.
    alt: object
        An optional module to be used in place of the given module if it
        fails to import in a non-GPU-enabled install

    Returns
    -------
    object
        The imported module, the given alternate, or a class derived from
        UnavailableMeta.
    """

    return safe_import(
        module,
        msg=f"{module} is not enabled in non GPU-enabled installations or environments. {GPU_INSTALL_STRING}",
        alt=alt,
    )


def gpu_only_import_from(module, symbol, *, alt=None):
    """A function used to import symbols required only in GPU installs

    This function will attempt to import a module with the given name.
    This function will attempt to import a symbol with the given name from
    the given module, but it will not throw an ImportError if the symbol is not
    found. Instead, it will return a placeholder object which will raise an
    exception only if used with instructions on installing a GPU build.

    Parameters
    ----------
    module: str
        The name of the module to import.
    symbol: str
        The name of the symbol to import.
    alt: object
        An optional object to be used in place of the given symbol if it fails
        to import in a non-GPU-enabled install

    Returns
    -------
    object
        The imported symbol, the given alternate, or a class derived from
        UnavailableMeta.
    """
    return safe_import_from(
        module,
        symbol,
        msg=f"{module}.{symbol} is not enabled in non GPU-enabled installations or environments. {GPU_INSTALL_STRING}",
        alt=alt,
    )


IMAGE_INSTALL_STRING = """Install image packages via `pip install --extra-index-url https://pypi.nvidia.com nemo-curator[image]`
or use `pip install --extra-index-url https://pypi.nvidia.com ".[image]"` if installing from source"""


def image_only_import_from(module, symbol, *, alt=None):
    """A function used to import symbols required only in image installs

    This function will attempt to import a module with the given name.
    This function will attempt to import a symbol with the given name from
    the given module, but it will not throw an ImportError if the symbol is not
    found. Instead, it will return a placeholder object which will raise an
    exception only if used with instructions on installing an image build.

    Parameters
    ----------
    module: str
        The name of the module to import.
    symbol: str
        The name of the symbol to import.
    alt: object
        An optional object to be used in place of the given symbol if it fails
        to import in a non-image install

    Returns
    -------
    object
        The imported symbol, the given alternate, or a class derived from
        UnavailableMeta.
    """
    return safe_import_from(
        module,
        symbol,
        msg=f"{module}.{symbol} is not enabled in without the nemo-curator[image] installations or environments. {IMAGE_INSTALL_STRING}",
        alt=alt,
    )



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

import nemo_curator as nc
from nemo_curator.datasets import DocumentDataset
from nemo_curator.tasks import (
    ANLI,
    CB,
    PIQA,
    RTE,
    WSC,
    ArcChallenge,
    ArcEasy,
    BoolQ,
    Copa,
    Drop,
    MultiRC,
    OpenBookQA,
    Quac,
    Race,
    Record,
    Squad,
    TriviaQA,
    WebQA,
    WiC,
    Winogrande,
)
from nemo_curator.utils.distributed_utils import get_client, read_data, write_to_disk
from nemo_curator.utils.file_utils import get_all_files_paths_under
from nemo_curator.utils.script_utils import ArgumentHelper


def load_dataset(input_data_dir):
    files = list(get_all_files_paths_under(input_data_dir, keep_extensions="jsonl"))
    raw_data = read_data(files, file_type="jsonl", backend="pandas", add_filename=True)
    dataset = DocumentDataset(raw_data)

    return dataset


def main(args):
    # Params
    contaminated_dataset_path = "/path/to/input"
    decontaminated_output_path = "/path/to/output"

    downstream_tasks = [
        Winogrande(),
        Squad(),
        TriviaQA(),
        Quac(),
        WebQA(),
        Race(),
        Drop(),
        WiC(),
        PIQA(),
        ArcEasy(),
        ArcChallenge(),
        OpenBookQA(),
        BoolQ(),
        Copa(),
        RTE(),
        MultiRC(),
        WSC(),
        CB(),
        ANLI(),
        Record(),
    ]

    # Prepare samples for the classifier
    client = get_client(**ArgumentHelper.parse_client_args(args))

    # Filter data
    target_dataset = load_dataset(contaminated_dataset_path)
    decontaminator = nc.TaskDecontamination(downstream_tasks)
    decontaminated_dataset = decontaminator(target_dataset)

    # Write filtered dataset
    write_to_disk(
        decontaminated_dataset.df, decontaminated_output_path, write_to_filename=True
    )


def attach_args(
    parser=argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    ),
):
    return ArgumentHelper(parser).add_distributed_args()


if __name__ == "__main__":
    main(attach_args().parse_args())



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from functools import wraps
from typing import Any, Callable, List, Literal, Optional, Union

import dask.dataframe as dd

from nemo_curator.utils.distributed_utils import read_data, write_to_disk
from nemo_curator.utils.file_utils import get_all_files_paths_under


class DocumentDataset:
    """
    A collection of documents and document metadata.
    Internally it may be distributed across multiple nodes, and may be on GPUs.
    """

    def __init__(self, dataset_df: dd.DataFrame):
        self.df = dataset_df

    def __len__(self) -> int:
        return len(self.df)

    # `def persist(self) -> Self` requires Python 3.11 or higher
    def persist(self) -> "DocumentDataset":
        return DocumentDataset(self.df.persist())

    @wraps(dd.DataFrame.repartition)
    def repartition(self, *args, **kwargs) -> "DocumentDataset":
        return self.__class__(self.df.repartition(*args, **kwargs))

    def head(self, n: int = 5) -> Any:
        return self.df.head(n)

    @classmethod
    def read_json(
        cls,
        input_files: Union[str, List[str]],
        backend: Literal["pandas", "cudf"] = "pandas",
        files_per_partition: Optional[int] = None,
        blocksize: Optional[str] = "1gb",
        add_filename: Union[bool, str] = False,
        input_meta: Optional[Union[str, dict]] = None,
        columns: Optional[List[str]] = None,
        **kwargs,
    ) -> "DocumentDataset":
        """
        Read JSONL or JSONL file(s).

        Args:
            input_files: The path of the input file(s).
            backend: The backend to use for reading the data.
            files_per_partition: The number of files to read per partition.
            add_filename: Whether to add a filename column to the DataFrame.
                If True, a new column is added to the DataFrame called `file_name`.
                If str, sets new column name. Default is False.
            input_meta: A dictionary or a string formatted as a dictionary, which outlines
                the field names and their respective data types within the JSONL input file.
            columns: If not None, only these columns will be read from the file.

        """
        return cls(
            _read_json_or_parquet(
                input_files=input_files,
                file_type="jsonl",
                backend=backend,
                add_filename=add_filename,
                files_per_partition=files_per_partition,
                blocksize=blocksize,
                input_meta=input_meta,
                columns=columns,
                **kwargs,
            )
        )

    @classmethod
    def read_parquet(
        cls,
        input_files: Union[str, List[str]],
        backend: Literal["pandas", "cudf"] = "pandas",
        files_per_partition: Optional[int] = None,
        blocksize: Optional[str] = "1gb",
        add_filename: Union[bool, str] = False,
        columns: Optional[List[str]] = None,
        **kwargs,
    ) -> "DocumentDataset":
        """
        Read Parquet file(s).

        Args:
            input_files: The path of the input file(s).
            backend: The backend to use for reading the data.
            files_per_partition: The number of files to read per partition.
            add_filename: Whether to add a filename column to the DataFrame.
                If True, a new column is added to the DataFrame called `file_name`.
                If str, sets new column name. Default is False.
            columns: If not None, only these columns will be read from the file.
                There is a significant performance gain when specifying columns for Parquet files.

        """
        return cls(
            _read_json_or_parquet(
                input_files=input_files,
                file_type="parquet",
                backend=backend,
                add_filename=add_filename,
                files_per_partition=files_per_partition,
                blocksize=blocksize,
                columns=columns,
                **kwargs,
            )
        )

    @classmethod
    def read_pickle(
        cls,
        input_files: Union[str, List[str]],
        backend: Literal["pandas", "cudf"] = "pandas",
        columns: Optional[List[str]] = None,
        **kwargs,
    ) -> "DocumentDataset":
        """
        Read Pickle file(s).

        Args:
            input_files: The path of the input file(s).
            backend: The backend to use for reading the data.
            files_per_partition: The number of files to read per partition.
            add_filename: Whether to add a filename column to the DataFrame.
                If True, a new column is added to the DataFrame called `file_name`.
                If str, sets new column name. Default is False.
            columns: If not None, only these columns will be read from the file.

        """
        return cls(
            read_data(
                input_files=input_files,
                file_type="pickle",
                backend=backend,
                columns=columns,
                **kwargs,
            )
        )

    @classmethod
    def read_custom(
        cls,
        input_files: Union[str, List[str]],
        file_type: str,
        read_func_single_partition: Callable[
            [List[str], str, bool, Union[str, dict], dict],
            Union["cudf.DataFrame", "pd.DataFrame"],
        ],
        files_per_partition: Optional[int] = None,
        backend: Optional[Literal["pandas", "cudf"]] = None,
        add_filename: Union[bool, str] = False,
        columns: Optional[List[str]] = None,
        input_meta: Union[str, dict] = None,
        **kwargs,
    ) -> "DocumentDataset":
        """
        Read custom data from a file or directory based on a custom read function.


        Args:
            input_files: The path of the input file(s).
                If input_file is a string that ends with the file_type, we consider it as a single file.
                If input_file is a string that does not end with the file_type, we consider it as a directory
                and read all files under the directory.
                If input_file is a list of strings, we assume each string is a file path.
            file_type: The type of the file to read.
            read_func_single_partition: A function that reads a single file or a list of files in an single dask partition.
                The function should take the following arguments:
                - files: A list of file paths.
                - file_type: The type of the file to read (in case you want to handle different file types differently).
                - backend: Read below
                - add_filename: Read below
                - columns: Read below
                - input_meta: Read below
            backend: The backend to use for reading the data, in case you want to handle pd.DataFrame or cudf.DataFrame.
            files_per_partition: The number of files to read per partition.
            add_filename: Whether to add a filename column to the DataFrame.
                If True, a new column is added to the DataFrame called `file_name`.
                If str, sets new column name. Default is False.
            columns: If not None, only these columns will be returned from the output of the read_func_single_partition function.
            input_meta: A dictionary or a string formatted as a dictionary, which outlines
                the field names and their respective data types within the JSONL input file.
        """
        if isinstance(input_files, str):
            if input_files.endswith(file_type):
                files = [input_files]
            else:
                files = get_all_files_paths_under(
                    root=input_files,
                    recurse_subdirectories=False,
                    keep_extensions=[file_type],
                )
        elif isinstance(input_files, list):
            files = input_files
        else:
            raise TypeError("input_files must be a string or list")
        return cls(
            read_data(
                input_files=files,
                backend=backend,
                files_per_partition=files_per_partition,
                blocksize=None,
                add_filename=add_filename,
                columns=columns,
                input_meta=input_meta,
                read_func_single_partition=read_func_single_partition,
                **kwargs,
            )
        )

    def to_json(
        self,
        output_path: str,
        write_to_filename: Union[bool, str] = False,
        keep_filename_column: bool = False,
        partition_on: Optional[str] = None,
        compression: Optional[str] = None,
    ):
        """
        Writes the dataset to the specified path in JSONL format.

        If `write_to_filename` is True, the DataFrame is expected to have a column
        that specifies the filename for each document. This column can be named
        `file_name` by default, or a custom name if `write_to_filename` is a string.

        Args:
            output_path (str): The directory or file path where the dataset will be written.
            write_to_filename (Union[bool, str]): Determines how filenames are handled.
                - If True, uses the `file_name` column in the DataFrame to determine filenames.
                - If a string, uses that string as the column name for filenames.
                - If False, writes all data to the specified `output_path`.
            keep_filename_column (bool): If True, retains the filename column in the output.
                If False, the filename column is dropped from the output.
            partition_on (Optional[str]): The column name used to partition the data.
                If specified, data is partitioned based on unique values in this column,
                with each partition written to a separate directory.
            compression (Optional[str]): The compression to use for the output file.
                If specified, the output file will be compressed using the specified compression.
                Supported compression types are "gzip" or None.
        For more details, refer to the `write_to_disk` function in
        `nemo_curator.utils.distributed_utils`.
        """
        write_to_disk(
            df=self.df,
            output_path=output_path,
            write_to_filename=write_to_filename,
            keep_filename_column=keep_filename_column,
            partition_on=partition_on,
            output_type="jsonl",
            compression=compression,
        )

    def to_parquet(
        self,
        output_path: str,
        write_to_filename: Union[bool, str] = False,
        keep_filename_column: bool = False,
        partition_on: Optional[str] = None,
    ):
        """
        Writes the dataset to the specified path in Parquet format.

        If `write_to_filename` is True, the DataFrame is expected to have a column
        that specifies the filename for each document. This column can be named
        `file_name` by default, or a custom name if `write_to_filename` is a string.

        Args:
            output_path (str): The directory or file path where the dataset will be written.
            write_to_filename (Union[bool, str]): Determines how filenames are handled.
                - If True, uses the `file_name` column in the DataFrame to determine filenames.
                - If a string, uses that string as the column name for filenames.
                - If False, writes all data to the specified `output_path`.
            keep_filename_column (bool): If True, retains the filename column in the output.
                If False, the filename column is dropped from the output.
            partition_on (Optional[str]): The column name used to partition the data.
                If specified, data is partitioned based on unique values in this column,
                with each partition written to a separate directory.

        For more details, refer to the `write_to_disk` function in
        `nemo_curator.utils.distributed_utils`.
        """
        write_to_disk(
            df=self.df,
            output_path=output_path,
            write_to_filename=write_to_filename,
            keep_filename_column=keep_filename_column,
            partition_on=partition_on,
            output_type="parquet",
        )

    def to_pickle(
        self,
        output_path: str,
        write_to_filename: Union[bool, str] = False,
    ):
        raise NotImplementedError("DocumentDataset does not support to_pickle yet")

    @classmethod
    def from_pandas(
        cls,
        data,
        npartitions: Optional[int] = 1,
        chunksize: Optional[int] = None,
        sort: Optional[bool] = True,
        name: Optional[str] = None,
    ):
        """
        Creates a document dataset from a pandas data frame.
        For more information on the arguments see Dask's from_pandas documentation
        https://docs.dask.org/en/stable/generated/dask.dataframe.from_pandas.html

        Args:
            data: A pandas dataframe
        Returns:
            A document dataset with a pandas backend (on the CPU).
        """
        return cls(
            dd.from_pandas(
                data=data,
                npartitions=npartitions,
                chunksize=chunksize,
                sort=sort,
            )
        )

    def to_pandas(self):
        """
        Creates a pandas dataframe from a DocumentDataset

        Returns:
            A pandas dataframe (on the CPU)
        """
        return self.df.to_backend("pandas").compute()


def _read_json_or_parquet(
    input_files: Union[str, List[str]],
    file_type: str,
    backend: Literal["cudf", "pandas"],
    add_filename: Union[bool, str] = False,
    files_per_partition: Optional[int] = None,
    blocksize: Optional[str] = None,
    input_meta: Union[str, dict] = None,
    columns: Optional[List[str]] = None,
    **kwargs,
):
    """
    `input_files` may be a list or a string type.
    If `input_files` is a list, it may be a list of JSONL or Parquet files,
    e.g., `input_files = ["file1.jsonl", "file2.jsonl", "file3.jsonl"]`,
    or a list of directories containing JSONL or Parquet files,
    e.g., `input_files = ["dir1", "dir2", "dir3"]`,
    where each of `dir1`, `dir2`, and `dir3` contain all JSONL or Parquet files.
    If `input_files` is a string, it may be a single JSONL or Parquet file,
    such as `input_files = "my_file.jsonl"`,
    or it may also be a single directory containing JSONL or Parquet files,
    such as `input_files = "my_directory"`.

    See nemo_curator.utils.distributed_utils.read_data docstring for other parameters.

    Returns a DataFrame to be used in initializing a DocumentDataset.

    """
    file_ext = "." + file_type

    if isinstance(input_files, list):
        # List of files
        if all(os.path.isfile(f) for f in input_files):
            raw_data = read_data(
                input_files,
                file_type=file_type,
                backend=backend,
                files_per_partition=files_per_partition,
                blocksize=blocksize,
                add_filename=add_filename,
                input_meta=input_meta,
                columns=columns,
                **kwargs,
            )

        # List of directories
        else:
            dfs = []

            for data_path in input_files:
                files = get_all_files_paths_under(
                    root=data_path, recurse_subdirectories=False
                )
                df = read_data(
                    files,
                    file_type=file_type,
                    backend=backend,
                    files_per_partition=files_per_partition,
                    blocksize=blocksize,
                    add_filename=add_filename,
                    input_meta=input_meta,
                    columns=columns,
                    **kwargs,
                )
                dfs.append(df)

            raw_data = dd.concat(dfs, ignore_unknown_divisions=True)

    elif isinstance(input_files, str):
        # Single file
        if input_files.endswith(file_ext):
            files = [input_files]

        # Directory of jsonl or parquet files
        else:
            files = get_all_files_paths_under(
                root=input_files, recurse_subdirectories=False
            )

        raw_data = read_data(
            input_files=files,
            file_type=file_type,
            backend=backend,
            files_per_partition=files_per_partition,
            blocksize=blocksize,
            add_filename=add_filename,
            input_meta=input_meta,
            columns=columns,
            **kwargs,
        )

    else:
        raise TypeError("File input must be a string or list.")

    return raw_data



import json
import logging
import os

import dask.dataframe as dd
import pandas as pd
from dask.distributed import Client, LocalCluster

from nemo_curator.utils.distributed_utils import get_num_workers
from nemo_curator.utils.module_utils import count_digits

logging.basicConfig(format="%(asctime)s: %(message)s", level=logging.INFO)

DATA_BASE = os.environ.get("DATA_BASE")
RAW_DATA_BASE = os.path.join(DATA_BASE, "processed")
CC_BASE = os.path.join(DATA_BASE, "fuzzy/cc/")
CC_FOLDER = os.path.join(CC_BASE, "connected_components.parquet")
CC_GROUPED_COUNTS_FOLDER = os.path.join(
    CC_BASE, "connected_components_grouped_counts.parquet"
)

DUPES_BASE = os.path.join(CC_BASE, "dupes")
DUPES_IDS_GROUPED_IN_COLUMNS = os.path.join(
    DUPES_BASE, "dupes_ids_grouped_in_columns.parquet"
)

CPU_WORKERS = os.environ.get("CPU_WORKERS")


if __name__ == "__main__":
    cluster = LocalCluster(n_workers=CPU_WORKERS, processes=True)
    client = Client(cluster)
    logging.info(f"Number of dask workers: {get_num_workers(client)}")

    paths = {
        "dclm": os.path.join(RAW_DATA_BASE, "dclm-baseline-1.0-parquet/filtered"),
        "dolma-cc": os.path.join(RAW_DATA_BASE, "dolma-v1_7-cc-parquet"),
        "fwe2": os.path.join(RAW_DATA_BASE, "fineweb-edu-score-2/data"),
        "zyda": os.path.join(RAW_DATA_BASE, "data/zyda_no_starcoder"),
    }

    dclm_id2dir = {
        "gs1": "global-shard_01_of_10",
        "gs2": "global-shard_02_of_10",
        "gs3": "global-shard_03_of_10",
        "gs4": "global-shard_04_of_10",
        "gs5": "global-shard_05_of_10",
        "gs6": "global-shard_06_of_10",
        "gs7": "global-shard_07_of_10",
        "gs8": "global-shard_08_of_10",
        "gs9": "global-shard_09_of_10",
        "gs10": "global-shard_10_of_10",
    }

    dclm_dir2id = {}
    for key, val in dclm_id2dir.items():
        dclm_dir2id[val] = key

    # Counting digits
    dclm_digits = {}
    for dir in sorted(os.listdir(paths["dclm"])):
        files = [
            x for x in os.listdir(os.path.join(paths["dclm"], dir)) if ".parquet" in x
        ]
        dclm_digits[dclm_dir2id[dir]] = count_digits(len(files))

    dolma_digits = count_digits(
        len([x for x in os.listdir(paths["dolma-cc"]) if ".parquet" in x])
    )

    zyda_digits = {}
    for dir in sorted(os.listdir(paths["zyda"])):
        files = [
            x for x in os.listdir(os.path.join(paths["zyda"], dir)) if ".parquet" in x
        ]
        zyda_digits[dir] = count_digits(len(files))

    fwe2_digits = {}
    for dir in sorted(os.listdir(paths["fwe2"])):
        files = [
            x for x in os.listdir(os.path.join(paths["fwe2"], dir)) if ".parquet" in x
        ]
        fwe2_digits[dir] = count_digits(len(files))

    cc_grouped_counts_df = dd.read_parquet(
        CC_GROUPED_COUNTS_FOLDER, split_row_groups=False
    )
    cc_grouped_counts_filtered_df = cc_grouped_counts_df[
        cc_grouped_counts_df["size"] > 1
    ]

    cc_groups_counts_inter_df = cc_grouped_counts_filtered_df[
        cc_grouped_counts_filtered_df["size"] != cc_grouped_counts_filtered_df["dclm"]
    ]
    cc_groups_counts_inter_df = cc_groups_counts_inter_df[
        cc_groups_counts_inter_df["size"] != cc_groups_counts_inter_df["fwe2"]
    ]
    cc_groups_counts_inter_df = cc_groups_counts_inter_df[
        cc_groups_counts_inter_df["size"] != cc_groups_counts_inter_df["dolma-cc"]
    ]
    cc_groups_counts_inter_df = cc_groups_counts_inter_df[
        cc_groups_counts_inter_df["size"] != cc_groups_counts_inter_df["zyda"]
    ]

    def select_dupes(partition):
        # Removes all overlaps with fwe2
        partition_fwe2 = partition[partition["fwe2"] > 0]
        partition_fwe2["dupes_to_remove"] = partition_fwe2["original_id"].apply(
            lambda x: json.dumps([id for id in json.loads(x) if "fwe2" not in id])
        )

        # Removes all overlaps with fwe2 (after fwe2 overlaps are removed)
        partition_dclm = partition[partition["fwe2"] == 0]
        partition_dclm = partition_dclm[partition_dclm["dclm"] > 0]
        partition_dclm["dupes_to_remove"] = partition_dclm["original_id"].apply(
            lambda x: json.dumps([id for id in json.loads(x) if "dclm" not in id])
        )

        # Removes all overlaps with zyda (after dclm, fwe2 overlaps are removed)
        partition_zyda = partition[partition["fwe2"] == 0]
        partition_zyda = partition_zyda[partition_zyda["dclm"] == 0]
        partition_zyda = partition_zyda[partition_zyda["zyda"] > 0]
        partition_zyda["dupes_to_remove"] = partition_zyda["original_id"].apply(
            lambda x: json.dumps([id for id in json.loads(x) if "zyda" not in id])
        )

        return pd.concat([partition_dclm, partition_fwe2, partition_zyda])

    meta = {
        "group": int,
        "global_dataset_id": str,
        "dataset_id": str,
        "original_id": str,
        "size": int,
        "dclm": int,
        "fwe2": int,
        "dolma-cc": int,
        "zyda": int,
        "dupes_to_remove": str,
    }
    dupes_df = cc_groups_counts_inter_df.map_partitions(select_dupes, meta=meta)

    def group_dupes(partition):
        partition["dclm_dupes"] = partition["dupes_to_remove"].apply(
            lambda x: json.dumps([id for id in json.loads(x) if "dclm" in id])
        )
        partition["zyda_dupes"] = partition["dupes_to_remove"].apply(
            lambda x: json.dumps([id for id in json.loads(x) if "zyda" in id])
        )
        partition["dolma_dupes"] = partition["dupes_to_remove"].apply(
            lambda x: json.dumps([id for id in json.loads(x) if "dolma" in id])
        )

        return partition[
            [
                "group",
                "size",
                "dclm",
                "fwe2",
                "dolma-cc",
                "zyda",
                "dclm_dupes",
                "zyda_dupes",
                "dolma_dupes",
            ]
        ]

    meta = {
        "group": int,
        "size": int,
        "dclm": int,
        "fwe2": int,
        "dolma-cc": int,
        "zyda": int,
        "dclm_dupes": str,
        "zyda_dupes": str,
        "dolma_dupes": str,
    }

    grouped_dupes_df = dupes_df.map_partitions(group_dupes, meta=meta)
    grouped_dupes_df.to_parquet(
        DUPES_IDS_GROUPED_IN_COLUMNS, write_index=False, overwrite=True
    )



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nemo_curator.filters import DocumentFilter


class FilterLowScores(DocumentFilter):
    """
    Discards documents with scores (human-assigned, or reward model assiegned) below a threshold.
    """

    def __init__(self, score_threshold: int):
        super().__init__()
        self._score_threshold = score_threshold

    def score_document(self, text: str) -> bool:
        return text >= self._score_threshold

    def keep_document(self, score) -> bool:
        return score



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
from typing import Any, List, Optional

from transformers import AutoTokenizer

from nemo_curator import BaseModule
from nemo_curator.datasets import DocumentDataset
from nemo_curator.services import LLMClient
from nemo_curator.synthetic.prompts import (
    DISTILL_PROMPT_TEMPLATE,
    DIVERSE_QA_PROMPT_TEMPLATE,
    EXTRACT_KNOWLEDGE_PROMPT_TEMPLATE,
    KNOWLEDGE_LIST_PROMPT_TEMPLATE,
    NEMOTRON_CC_DISTILL_SYSTEM_PROMPT,
    NEMOTRON_CC_SYSTEM_PROMPT,
    WIKIPEDIA_REPHRASING_PROMPT_TEMPLATE,
)


class NemotronCCGenerator:
    """
    Provides a collection of methods for generating synthetic data
    described in the Nemotron-CC paper (https://arxiv.org/abs/2412.02595).
    """

    def __init__(self, llm_client: LLMClient) -> None:
        """
        Initialize the NemotronCCGenerator instance.

        Args:
            llm_client (LLMClient): The language model client used for querying the model.
        """
        self.client = llm_client

    def _prompt(
        self,
        model: str,
        document: str,
        prompt_template: str,
        system_prompt: str,
        prompt_kwargs: dict,
        model_kwargs: dict,
    ) -> List[str]:
        prompt = prompt_template.format(document=document, **prompt_kwargs)
        messages = [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": prompt},
        ]

        return self.client.query_model(messages=messages, model=model, **model_kwargs)

    def rewrite_to_wikipedia_style(
        self,
        document: str,
        model: str,
        prompt_template: str = WIKIPEDIA_REPHRASING_PROMPT_TEMPLATE,
        system_prompt: str = NEMOTRON_CC_SYSTEM_PROMPT,
        prompt_kwargs: dict = {},
        model_kwargs: dict = {},
    ) -> List[str]:
        """
        Rewrites a document into a Wikipedia-style narrative.

        Args:
            document (str): The input document text to rewrite.
            model (str): The model identifier to use.
            prompt_template (str, optional): The prompt template for rewriting. Defaults to WIKIPEDIA_REPHRASING_PROMPT_TEMPLATE.
            system_prompt (str, optional): The system prompt to use. Defaults to NEMOTRON_CC_SYSTEM_PROMPT.
            prompt_kwargs (dict, optional): Additional keyword arguments for the prompt. Defaults to {}.
            model_kwargs (dict, optional): Additional keyword arguments for the model invocation. Defaults to {}.

        Returns:
            List[str]: A list of responses from the LLM. The list is only greater than length 1 if n > 1 is set in model_kwargs.
        """
        return self._prompt(
            model, document, prompt_template, system_prompt, prompt_kwargs, model_kwargs
        )

    def generate_diverse_qa(
        self,
        document: str,
        model: str,
        prompt_template: str = DIVERSE_QA_PROMPT_TEMPLATE,
        system_prompt: str = NEMOTRON_CC_SYSTEM_PROMPT,
        prompt_kwargs: dict = {},
        model_kwargs: dict = {},
    ) -> List[str]:
        """
        Generates diverse QA pairs from the provided document.

        Args:
            document (str): The input document text used to generate QA pairs.
            model (str): The model identifier to use.
            prompt_template (str, optional): The prompt template for generating QA pairs. Defaults to DIVERSE_QA_PROMPT_TEMPLATE.
            system_prompt (str, optional): The system prompt to use. Defaults to NEMOTRON_CC_SYSTEM_PROMPT.
            prompt_kwargs (dict, optional): Additional keyword arguments for the prompt. Defaults to {}.
            model_kwargs (dict, optional): Additional keyword arguments for the model invocation. Defaults to {}.

        Returns:
            List[str]: A list of responses from the LLM. The list is only greater than length 1 if n > 1 is set in model_kwargs.
        """
        return self._prompt(
            model, document, prompt_template, system_prompt, prompt_kwargs, model_kwargs
        )

    def distill(
        self,
        document: str,
        model: str,
        prompt_template: str = DISTILL_PROMPT_TEMPLATE,
        system_prompt: str = NEMOTRON_CC_DISTILL_SYSTEM_PROMPT,
        prompt_kwargs: dict = {},
        model_kwargs: dict = {},
    ) -> List[str]:
        """
        Distills the essential content from a document.

        Args:
            document (str): The input document text to distill.
            model (str): The model identifier to use.
            prompt_template (str, optional): The prompt template for distillation. Defaults to DISTILL_PROMPT_TEMPLATE.
            system_prompt (str, optional): The system prompt to use. Defaults to NEMOTRON_CC_DISTILL_SYSTEM_PROMPT.
            prompt_kwargs (dict, optional): Additional keyword arguments for the prompt. Defaults to {}.
            model_kwargs (dict, optional): Additional keyword arguments for the model invocation. Defaults to {}.

        Returns:
            List[str]: A list of responses from the LLM. The list is only greater than length 1 if n > 1 is set in model_kwargs.
        """
        return self._prompt(
            model, document, prompt_template, system_prompt, prompt_kwargs, model_kwargs
        )

    def extract_knowledge(
        self,
        document: str,
        model: str,
        prompt_template: str = EXTRACT_KNOWLEDGE_PROMPT_TEMPLATE,
        system_prompt: str = NEMOTRON_CC_SYSTEM_PROMPT,
        prompt_kwargs: dict = {},
        model_kwargs: dict = {},
    ) -> List[str]:
        """
        Extracts knowledge from the provided document.

        Args:
            document (str): The input document text from which to extract knowledge.
            model (str): The model identifier to use.
            prompt_template (str, optional): The prompt template for knowledge extraction. Defaults to EXTRACT_KNOWLEDGE_PROMPT_TEMPLATE.
            system_prompt (str, optional): The system prompt to use. Defaults to NEMOTRON_CC_SYSTEM_PROMPT.
            prompt_kwargs (dict, optional): Additional keyword arguments for the prompt. Defaults to {}.
            model_kwargs (dict, optional): Additional keyword arguments for the model invocation. Defaults to {}.

        Returns:
            List[str]: A list of responses from the LLM. The list is only greater than length 1 if n > 1 is set in model_kwargs.
        """
        return self._prompt(
            model, document, prompt_template, system_prompt, prompt_kwargs, model_kwargs
        )

    def generate_knowledge_list(
        self,
        document: str,
        model: str,
        prompt_template: str = KNOWLEDGE_LIST_PROMPT_TEMPLATE,
        system_prompt: str = NEMOTRON_CC_SYSTEM_PROMPT,
        prompt_kwargs: dict = {},
        model_kwargs: dict = {},
    ) -> List[str]:
        """
        Generates a list of knowledge items from the provided document.

        Args:
            document (str): The input document text to process.
            model (str): The model identifier to use.
            prompt_template (str, optional): The prompt template for generating a knowledge list. Defaults to KNOWLEDGE_LIST_PROMPT_TEMPLATE.
            system_prompt (str, optional): The system prompt to use. Defaults to NEMOTRON_CC_SYSTEM_PROMPT.
            prompt_kwargs (dict, optional): Additional keyword arguments for the prompt. Defaults to {}.
            model_kwargs (dict, optional): Additional keyword arguments for the model invocation. Defaults to {}.

        Returns:
            List[str]: A list of responses from the LLM. The list is only greater than length 1 if n > 1 is set in model_kwargs.
        """
        return self._prompt(
            model, document, prompt_template, system_prompt, prompt_kwargs, model_kwargs
        )


class NemotronCCDiverseQAPostprocessor(BaseModule):
    """
    Postprocesses the output of the Nemotron-CC Diverse QA generation pipeline.
    This postprocessor will sample a random number of QA pairs up to max_num_pairs.
    If a tokenizer is provided, the number of QA pairs will be sampled from at least 1 and at most floor(max_num_pairs * num_tokens / 150).
    Otherwise, the number of QA pairs will be sampled randomly strictly up to max_num_pairs.

    The generated QA pairs are shuffled and then appended to the original text.
    """

    def __init__(
        self,
        tokenizer: Optional[AutoTokenizer] = None,
        text_field: str = "text",
        response_field: str = "response",
        max_num_pairs: int = 1,
        prefix: str = "Here are the questions and answers based on the provided text:",
    ) -> None:
        """
        Args:
            tokenizer (Optional[AutoTokenizer]): The tokenizer to use for tokenization.
                If specified, the number of QA pairs will be sampled based on the token count of the text.
                If not specified, the number of QA pairs will be sampled randomly up to max_num_pairs.
            text_field (str): The field in the dataset that contains the text used to generate QA pairs.
            response_field (str): The field in the dataset that contains the response from the LLM.
            max_num_pairs (int): The maximum number of QA pairs to sample.
            prefix (str): The prefix of the response from the LLM.
        """
        super().__init__(input_backend="pandas")
        self.tokenizer = tokenizer
        self.text_field = text_field
        self.response_field = response_field
        self.max_num_pairs = max_num_pairs
        self.prefix = prefix

    def _postprocess_llm_response(self, text: str, llm_response: str) -> str:
        lines = [line.strip() for line in llm_response.split("\n") if line.strip()]
        if not lines:
            return ""

        # Remove the "- " prefix
        lines = [line[2:].strip() if line.startswith("- ") else line for line in lines]

        if lines[0] == self.prefix:
            lines = lines[1:]

        # Merge question and answer lines
        qa_pairs = []
        for line in lines:
            if line.startswith("Question:"):
                qa_pairs.append(line)
            else:
                if qa_pairs:
                    qa_pairs[-1] += "\n" + line
                else:
                    return ""

        if len(qa_pairs) == 0:
            return ""

        # Shuffle the QA pairs and sample up to max_num_pairs
        random.shuffle(qa_pairs)
        if self.tokenizer is not None:
            num_tokens = len(self.tokenizer.tokenize(text))
            qa_pairs = qa_pairs[
                : random.randint(1, max(1, int(self.max_num_pairs * num_tokens / 150)))
            ]
        else:
            qa_pairs = qa_pairs[: random.randint(1, self.max_num_pairs)]
        qa_pairs_str = "\n\n".join(qa_pairs)

        # Concatenate the document and the QA pairs
        return f"{text}\n\n{qa_pairs_str}"

    def call(self, dataset: DocumentDataset) -> DocumentDataset:
        df = dataset.df
        df[self.response_field] = df.apply(
            lambda row: self._postprocess_llm_response(
                row[self.text_field], row[self.response_field]
            ),
            axis=1,
            meta=(None, "object"),
        )
        df = df[df[self.response_field] != ""]

        return DocumentDataset(df)


# Although this could be implemented as a DocumentModifier,
# I have kept it separate to match the other postprocessors.
class NemotronCCKnowledgeListPostprocessor(BaseModule):
    """
    Processes and cleans the output generated by the Nemotron-CC Knowledge List pipeline.

    This class is responsible for postprocessing raw text responses produced by the
    Nemotron-CC Knowledge List generation pipeline. It removes formatting artifacts
    such as bullet point prefixes ("- ") and extra indentation from each line, ensuring
    that the final output is a clean, uniformly formatted list of knowledge items.
    The processing includes skipping any initial non-bullet line and merging related lines
    to reconstruct multi-line questions or answers.
    """

    def __init__(self, text_field: str = "text") -> None:
        super().__init__(input_backend="pandas")
        self.text_field = text_field

    def _postprocess_llm_response(self, text: str) -> str:
        lines = []
        for idx, line in enumerate(text.split("\n")):
            if idx == 0 and not line.startswith("-"):
                continue

            if line.startswith("  ") or line.startswith("- "):
                lines.append(line[2:].strip())
            else:
                lines.append(line)
        return "\n".join(lines)

    def call(self, dataset: DocumentDataset) -> DocumentDataset:
        df = dataset.df
        df[self.text_field] = df[self.text_field].apply(
            self._postprocess_llm_response, meta=(self.text_field, "object")
        )
        return DocumentDataset(df)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dask.dataframe as dd
import pandas as pd
from dask.dataframe.utils import assert_eq

from nemo_curator import Modify
from nemo_curator.datasets import DocumentDataset
from nemo_curator.modifiers import (
    LineRemover,
    MarkdownRemover,
    NewlineNormalizer,
    QuotationRemover,
    Slicer,
    UnicodeReformatter,
    UrlRemover,
)


def list_to_dataset(documents, col_name="text", npartitions=2):
    data = {col_name: documents}
    pdf = pd.DataFrame(data)

    return DocumentDataset.from_pandas(pdf, npartitions=npartitions)


class TestUnicodeReformatter:
    def test_reformatting(self):
        # Examples taken from ftfy documentation:
        # https://ftfy.readthedocs.io/en/latest/
        dataset = list_to_dataset(
            [
                "âœ” No problems",
                "The Mona Lisa doesnÃƒÂ¢Ã¢â€šÂ¬Ã¢â€žÂ¢t have eyebrows.",
                "l’humanitÃ©",
                "Ã perturber la rÃ©flexion",
                "Clean document already.",
            ]
        )
        expected_results = [
            "✔ No problems",
            "The Mona Lisa doesn't have eyebrows.",
            "l'humanité",
            "à perturber la réflexion",
            "Clean document already.",
        ]
        expected_results.sort()

        modifier = Modify(UnicodeReformatter(uncurl_quotes=True))
        fixed_dataset = modifier(dataset)
        actual_results = fixed_dataset.df.compute()["text"].to_list()
        actual_results.sort()

        assert (
            expected_results == actual_results
        ), f"Expected: {expected_results}, but got: {actual_results}"


class TestNewlineNormalizer:
    def test_just_newlines(self):
        dataset = list_to_dataset(
            [
                "The quick brown fox jumps over the lazy dog",
                "The quick\nbrown fox jumps \nover the lazy dog",
                "The quick\n\nbrown fox jumps \n\nover the lazy dog",
                "The quick\n\n\nbrown fox jumps \n\n\nover the lazy dog",
                "The quick\n\n\nbrown fox jumps \nover the lazy dog",
            ]
        )
        expected_results = [
            "The quick brown fox jumps over the lazy dog",
            "The quick\nbrown fox jumps \nover the lazy dog",
            "The quick\n\nbrown fox jumps \n\nover the lazy dog",
            "The quick\n\nbrown fox jumps \n\nover the lazy dog",
            "The quick\n\nbrown fox jumps \nover the lazy dog",
        ]
        expected_results.sort()

        modifier = Modify(NewlineNormalizer())
        fixed_dataset = modifier(dataset)
        actual_results = fixed_dataset.df.compute()["text"].to_list()
        actual_results.sort()

        assert (
            expected_results == actual_results
        ), f"Expected: {expected_results}, but got: {actual_results}"

    def test_newlines_and_carriage_returns(self):
        dataset = list_to_dataset(
            [
                "The quick brown fox jumps over the lazy dog",
                "The quick\r\nbrown fox jumps \r\nover the lazy dog",
                "The quick\r\n\r\nbrown fox jumps \r\n\r\nover the lazy dog",
                "The quick\r\n\r\n\r\nbrown fox jumps \r\n\r\n\r\nover the lazy dog",
                "The quick\r\n\r\n\r\nbrown fox jumps \r\nover the lazy dog",
            ]
        )
        expected_results = [
            "The quick brown fox jumps over the lazy dog",
            "The quick\r\nbrown fox jumps \r\nover the lazy dog",
            "The quick\r\n\r\nbrown fox jumps \r\n\r\nover the lazy dog",
            "The quick\r\n\r\nbrown fox jumps \r\n\r\nover the lazy dog",
            "The quick\r\n\r\nbrown fox jumps \r\nover the lazy dog",
        ]
        expected_results.sort()

        modifier = Modify(NewlineNormalizer())
        fixed_dataset = modifier(dataset)
        actual_results = fixed_dataset.df.compute()["text"].to_list()
        actual_results.sort()

        assert (
            expected_results == actual_results
        ), f"Expected: {expected_results}, but got: {actual_results}"


class TestUrlRemover:
    def test_urls(self):
        dataset = list_to_dataset(
            [
                "This is a url: www.nvidia.com",
                "This is a url: http://www.nvidia.com",
                "This is a url: https://www.nvidia.com",
                "This is a url: https://www.nvidia.gov",
                "This is a url: https://nvidia.com",
                "This is a url: HTTPS://WWW.NVIDIA.COM",
                "This is not a url: git@github.com:NVIDIA/NeMo-Curator.git",
            ]
        )
        expected_results = [
            "This is a url: ",
            "This is a url: ",
            "This is a url: ",
            "This is a url: ",
            "This is a url: ",
            "This is a url: ",
            "This is not a url: git@github.com:NVIDIA/NeMo-Curator.git",
        ]
        expected_results.sort()

        modifier = Modify(UrlRemover())
        fixed_dataset = modifier(dataset)
        actual_results = fixed_dataset.df.compute()["text"].to_list()
        actual_results.sort()

        assert (
            expected_results == actual_results
        ), f"Expected: {expected_results}, but got: {actual_results}"


class TestLineRemover:
    def test_remove_exact_match(self):
        text = "Keep this\nRemove me\nAlso keep this\nRemove me"
        patterns = ["Remove me"]
        remover = LineRemover(patterns)
        result = remover.modify_document(text)
        expected = "Keep this\nAlso keep this"
        assert result == expected

    def test_no_removal_when_partial_match(self):
        text = (
            "Keep this line\nThis line contains Remove me as a part of it\nAnother line"
        )
        patterns = ["Remove me"]
        remover = LineRemover(patterns)
        # Only lines that exactly match "Remove me" are removed.
        assert remover.modify_document(text) == text

    def test_empty_input(self):
        text = ""
        patterns = ["Remove me"]
        remover = LineRemover(patterns)
        result = remover.modify_document(text)
        assert result == ""

    def test_multiple_patterns(self):
        text = "Line one\nDelete\nLine two\nRemove\nLine three\nDelete"
        patterns = ["Delete", "Remove"]
        remover = LineRemover(patterns)
        result = remover.modify_document(text)
        expected = "Line one\nLine two\nLine three"
        assert result == expected

    def test_whitespace_sensitivity(self):
        # Exact match requires identical string content.
        text = "Remove me \nRemove me\n  Remove me"
        patterns = ["Remove me"]
        remover = LineRemover(patterns)
        result = remover.modify_document(text)
        # Only the line that exactly equals "Remove me" is removed.
        expected = "Remove me \n  Remove me"
        assert result == expected

    def test_dataset_modification(self):
        docs = [
            "Keep this\nRemove me\nKeep that",
            "Remove me\nDon't remove\nRemove me",
            "No removal here",
            "Remove me",
        ]
        expected_results = [
            "Keep this\nKeep that",
            "Don't remove",
            "No removal here",
            "",
        ]
        dataset = list_to_dataset(docs)
        modifier = Modify(LineRemover(["Remove me"]))
        fixed_dataset = modifier(dataset)
        expected_dataset = list_to_dataset(expected_results)
        assert_eq(fixed_dataset.df, expected_dataset.df)


class TestQuotationRemover:
    def test_remove_quotes_no_newline(self):
        text = '"Hello, World!"'
        remover = QuotationRemover()
        result = remover.modify_document(text)
        expected = "Hello, World!"
        assert result == expected

    def test_no_removal_when_quotes_not_enclosing(self):
        text = 'Hello, "World!"'
        remover = QuotationRemover()
        result = remover.modify_document(text)
        # The text does not start and end with a quotation mark.
        assert result == text

    def test_remove_quotes_with_newline_removal(self):
        text = '"Hello,\nWorld!"'
        remover = QuotationRemover()
        result = remover.modify_document(text)
        # Since there is a newline and the first line does not end with a quote,
        # the quotes are removed.
        expected = "Hello,\nWorld!"
        assert result == expected

    def test_no_removal_with_newline_preserved(self):
        text = '"Hello,"\nWorld!"'
        remover = QuotationRemover()
        result = remover.modify_document(text)
        # The first line ends with a quote so the removal does not occur.
        assert result == text

    def test_short_text_no_removal(self):
        text = '""'
        remover = QuotationRemover()
        result = remover.modify_document(text)
        # With text length not greater than 2 (after stripping), nothing changes.
        assert result == text

    def test_extra_whitespace_prevents_removal(self):
        # If leading/trailing whitespace prevents the text from starting with a quote,
        # nothing is changed.
        text = '   "Test Message"   '
        remover = QuotationRemover()
        result = remover.modify_document(text)
        assert result == text

    def test_dataset_modification(self):
        import pandas as pd
        from dask.dataframe.utils import assert_eq

        docs = ['"Document one"', 'Start "Document two" End', '"Document\nthree"', '""']
        expected_results = [
            "Document one",
            'Start "Document two" End',
            "Document\nthree",
            '""',
        ]
        dataset = list_to_dataset(docs)
        modifier = Modify(QuotationRemover())
        fixed_dataset = modifier(dataset)
        expected_dataset = list_to_dataset(expected_results)
        assert_eq(fixed_dataset.df, expected_dataset.df)


class TestSlicer:
    def test_integer_indices(self):
        text = "Hello, world!"
        slicer = Slicer(left=7, right=12)
        result = slicer.modify_document(text)
        expected = "world"
        assert result == expected

    def test_left_string_including(self):
        text = "abcXYZdef"
        slicer = Slicer(left="XYZ", include_left=True)
        result = slicer.modify_document(text)
        expected = "XYZdef"
        assert result == expected

    def test_left_string_excluding(self):
        text = "abcXYZdef"
        slicer = Slicer(left="XYZ", include_left=False)
        result = slicer.modify_document(text)
        expected = "def"
        assert result == expected

    def test_right_string_including(self):
        text = "abcXYZdef"
        slicer = Slicer(right="XYZ", include_right=True)
        result = slicer.modify_document(text)
        expected = "abcXYZ"
        assert result == expected

    def test_right_string_excluding(self):
        text = "abcXYZdef"
        slicer = Slicer(right="XYZ", include_right=False)
        result = slicer.modify_document(text)
        expected = "abc"
        assert result == expected

    def test_both_left_and_right_with_strings(self):
        text = "start middle end"
        slicer = Slicer(
            left="start", right="end", include_left=False, include_right=False
        )
        result = slicer.modify_document(text)
        # "start" is removed and "end" is excluded; extra spaces are stripped.
        expected = "middle"
        assert result == expected

    def test_non_existing_left(self):
        text = "abcdef"
        slicer = Slicer(left="nonexistent")
        result = slicer.modify_document(text)
        assert result == ""

    def test_non_existing_right(self):
        text = "abcdef"
        slicer = Slicer(right="nonexistent")
        result = slicer.modify_document(text)
        assert result == ""

    def test_no_left_no_right(self):
        text = "   some text with spaces   "
        slicer = Slicer()
        result = slicer.modify_document(text)
        # With no boundaries specified, the entire text is returned (stripped).
        expected = "some text with spaces"
        assert result == expected

    def test_integer_out_of_range(self):
        text = "short"
        slicer = Slicer(left=10)
        result = slicer.modify_document(text)
        # Slicing starting beyond the text length yields an empty string.
        assert result == ""

    def test_multiple_occurrences(self):
        text = "abc__def__ghi"
        # Testing when markers appear multiple times.
        slicer = Slicer(left="__", right="__", include_left=True, include_right=True)
        result = slicer.modify_document(text)
        # left: first occurrence at index 3; right: last occurrence at index 8, include_right adds len("__")
        expected = "__def__"
        assert result == expected

    def test_dataset_modification(self):
        import pandas as pd
        from dask.dataframe.utils import assert_eq

        docs = ["abcdef", "0123456789", "Hello", "Slicer"]
        expected_results = [
            "cde",  # "abcdef" sliced from index 2 to 5
            "234",  # "0123456789" sliced from index 2 to 5
            "llo",  # "Hello" sliced from index 2 to 5
            "ice",  # "Slicer" sliced from index 2 to 5
        ]
        dataset = list_to_dataset(docs)
        modifier = Modify(Slicer(left=2, right=5))
        fixed_dataset = modifier(dataset)
        expected_dataset = list_to_dataset(expected_results)
        assert_eq(fixed_dataset.df, expected_dataset.df)


class TestMarkdownRemover:
    def test_bold_removal(self):
        text = "This is **bold** text."
        remover = MarkdownRemover()
        result = remover.modify_document(text)
        expected = "This is bold text."
        assert result == expected

    def test_italic_removal(self):
        text = "This is *italic* text."
        remover = MarkdownRemover()
        result = remover.modify_document(text)
        expected = "This is italic text."
        assert result == expected

    def test_underline_removal(self):
        text = "This is _underlined_ text."
        remover = MarkdownRemover()
        result = remover.modify_document(text)
        expected = "This is underlined text."
        assert result == expected

    def test_link_removal(self):
        text = "Link: [Google](https://google.com)"
        remover = MarkdownRemover()
        result = remover.modify_document(text)
        expected = "Link: https://google.com"
        assert result == expected

    def test_multiple_markdown(self):
        text = "This is **bold**, *italic*, and _underline_, check [Example](https://example.com)"
        remover = MarkdownRemover()
        result = remover.modify_document(text)
        expected = "This is bold, italic, and underline, check https://example.com"
        assert result == expected

    def test_no_markdown(self):
        text = "This line has no markdown."
        remover = MarkdownRemover()
        result = remover.modify_document(text)
        assert result == text

    def test_incomplete_markdown(self):
        text = "This is *italic text"
        remover = MarkdownRemover()
        result = remover.modify_document(text)
        # Without a closing '*', the text remains unchanged.
        assert result == text

    def test_nested_markdown(self):
        text = "This is **bold and *italic* inside** text."
        remover = MarkdownRemover()
        result = remover.modify_document(text)
        # Bold formatting is removed first, then italics in the resulting string.
        expected = "This is bold and italic inside text."
        assert result == expected

    def test_multiple_lines(self):
        text = "**Bold line**\n*Italic line*\n_Normal line_"
        remover = MarkdownRemover()
        result = remover.modify_document(text)
        expected = "Bold line\nItalic line\nNormal line"
        assert result == expected

    def test_adjacent_markdown(self):
        text = "**Bold****MoreBold**"
        remover = MarkdownRemover()
        result = remover.modify_document(text)
        expected = "BoldMoreBold"
        assert result == expected

    def test_dataset_modification(self):
        import pandas as pd
        from dask.dataframe.utils import assert_eq

        docs = [
            "This is **bold**",
            "This is *italic*",
            "Check [Link](https://example.com)",
            "No markdown here",
        ]
        expected_results = [
            "This is bold",
            "This is italic",
            "Check https://example.com",
            "No markdown here",
        ]
        dataset = list_to_dataset(docs)
        modifier = Modify(MarkdownRemover())
        fixed_dataset = modifier(dataset)
        expected_dataset = list_to_dataset(expected_results)
        assert_eq(fixed_dataset.df, expected_dataset.df)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gzip
import os
import re
from zipfile import ZipFile, ZipInfo

import arxiv as arxiv
import cchardet as chardet
import requests
from bs4 import BeautifulSoup
from unstructured.partition.auto import partition

from nemo_curator.download.doc_builder import (
    DocumentDownloader,
    DocumentExtractor,
    DocumentIterator,
)


class WikitxtDownloader(DocumentDownloader):
    """
    A class for downloading data from wiki URLs.
    """

    def __init__(self, download_dir: str):
        """
        Initializes the DocBuilder object.

        Args:
            download_dir: The root directory for wiki URLs
        """
        super().__init__()

        if not os.path.isdir(download_dir):
            os.makedirs(download_dir)

        self._download_dir = download_dir
        print("Download directory: ", self._download_dir)

    def download(self, url: str) -> str:
        """
        Download a url and extract content into a zip file.

        Args:
            url (str): The wiki URL.

        Returns:
            str: The path to the downloaded zip file, or None if the download failed.
        """
        # download wiki urls
        filename = os.path.basename(url)
        output_file = os.path.join(self._download_dir, filename) + ".wiki.gz"

        if os.path.exists(output_file):
            print(f"File '{output_file}' already exists, skipping download.")
            return output_file

        print(f"Downloading txt URLs data from '{url}'...")
        response = requests.get(url)
        html = BeautifulSoup(response.content, "lxml")
        title = html.select("#firstHeading")[0].text
        paragraphs = html.find_all(["p", "ul", "li"])
        intro = "\n".join([para.text for para in paragraphs])
        with gzip.open(output_file, "wt") as file:
            file.write(intro)
        return output_file


class WikitxtIterator(DocumentIterator):
    """
    Wiki document iterator. Will go through the files and parse.
    """

    # The token that separates paragraphs.
    SEPARATOR_TOKEN = "<|endoftext|>"

    def __init__(self):
        super().__init__()
        self._counter = -1

    def iterate(self, file_path):
        self._counter = -1
        file_name = os.path.basename(file_path)

        with gzip.open(file_path, "rt") as file:
            example = []

            def split_meta(example):
                if example:
                    self._counter += 1
                    content = " ".join(example)
                    line_count = content.count("  ") + 1
                    size_in_bytes = len(content.encode("utf-8"))
                    meta = {
                        "id": f"{file_name}-{self._counter}",
                        "file_extension": ".txt",
                        "file_type": "text",
                        "category": "text",
                        "line_count": line_count,
                        "size_in_bytes": size_in_bytes,
                        "path": file_name,
                    }

                    return meta, content

            for line in file:
                if line.strip() == WikitxtIterator.SEPARATOR_TOKEN:
                    if example:
                        yield split_meta(example)
                        example = []
                else:
                    example.append(line.strip())

            if example:
                yield split_meta(example)


class WikitxtExtractor(DocumentExtractor):
    def extract(self, content: str) -> dict:
        # No metadata for the text, just the content.
        return {"text": content}


class GitHubDownloader(DocumentDownloader):
    """
    A class for downloading repositories from GitHub.
    """

    def __init__(self, github_root_dir: str):
        """
        Initializes the DocBuilder object.

        Args:
            github_root_dir: The root directory for GitHub repositories.
        """
        super().__init__()
        # The path under which the repositories will be cloned.
        self.clone_root_dir = os.path.join(github_root_dir, "repos")
        os.makedirs(github_root_dir, exist_ok=True)
        os.makedirs(self.clone_root_dir, exist_ok=True)

    def download(self, url: str) -> str:
        """
        Download a repository as a zip file.

        Args:
            url (str): The URL of the repository.

        Returns:
            str: The path to the downloaded zip file, or None if the download failed.
        """
        repo_name = os.path.basename(url)
        zip_file = os.path.join(self.clone_root_dir, repo_name + ".zip")

        if os.path.exists(zip_file):
            print(f"Repository '{repo_name}' already exists, skipping download.")
            return zip_file

        # Try the common branch names first. A better way to do this would be to
        # query the GitHub API to get the default branch, but that is subject to rate limits.
        success = False

        for branch in ["master", "main"]:
            zip_url = f"https://github.com/{url}/archive/refs/heads/{branch}.zip"

            # Send a GET request to the URL
            response = requests.get(zip_url)

            # Check if the request was successful
            if response.status_code == 200:
                # Write the content of the response to a file
                with open(zip_file, "wb") as file:
                    file.write(response.content)

                # No need to try other branches
                success = True
                break

        if not success:
            print(
                f"Failed to clone repository '{repo_name}' from '{url}' (error code {response.status_code})."
            )
            return None

        return zip_file


class GitHubIterator(DocumentIterator):
    """
    GitHub document iterator. Will go through the files and parse the supported ones.
    """

    # Mapping from file extensions to categories.
    # Will also be used to to ignore irrelevant files.
    SUPPORTED_EXTENSIONS_TO_CATEGORY = {
        ".v": "VerilogVHDL",
        ".vh": "VerilogVHDL",
        ".vhdl": "VerilogVHDL",
        ".va": "VerilogAnalog",
        ".c": "CPP",
        ".cpp": "CPP",
        ".h": "CPP",
        ".hpp": "CPP",
        ".py": "Python",
        ".config": "Config",
        ".mk": "Makefile",
        "makefile": "Makefile",
        "makeppfile": "Makefile",
        ".pm": "Perl",
        ".pl": "Perl",
        ".tcl": "Tcl",
        ".spec": "Spec",
        ".yaml": "Yaml",
        ".yml": "Yaml",
        ".sp": "Spice",
        ".cir": "Spice",
        ".cmd": "Spice",
        ".spf": "Spice",
        ".spice": "Spice",
        ".txt": "text",
        ".json": "text",
        ".xml": "text",
        ".html": "text",
        ".pdf": "text",
        ".md": "text",
        "": "text",  # No extension
    }

    def parse_file(self, zip_ref: ZipFile, file_info: ZipInfo):
        """
        Parses a file from a zip archive and extracts its metadata and content.

        Args:
            zip_ref: The zip archive object.
            file_info: Information about the file in the zip archive.

        Returns:
            A tuple containing the metadata and the content of the file. The metadata is a dictionary.
            If the file extension or filename is not supported, or if the file cannot be decoded,
            None is returned.
        """
        zip_path = zip_ref.filename
        input_fp = file_info.filename
        full_path = os.path.join(zip_path, input_fp)
        # Extract the file name and extension in lower case.
        filename = os.path.basename(input_fp)
        filename_no_ext, ext = os.path.splitext(filename)
        filename_no_ext = filename_no_ext.lower()
        ext = ext.lower()

        # If neither the file extension nor the filename is supported, return None
        if ext not in GitHubIterator.SUPPORTED_EXTENSIONS_TO_CATEGORY:
            if filename_no_ext not in GitHubIterator.SUPPORTED_EXTENSIONS_TO_CATEGORY:
                return None

            # The filename is there, but the extension is not. The category is determined by the filename.
            category = GitHubIterator.SUPPORTED_EXTENSIONS_TO_CATEGORY[filename_no_ext]
        else:
            category = GitHubIterator.SUPPORTED_EXTENSIONS_TO_CATEGORY[ext]

        # Open the file and read its content. Determine the encoding using cchardet. Skip over binary files.
        with zip_ref.open(file_info, "r") as file:
            content = file.read()
            # Determine the encoding of the file
            encoding = chardet.detect(content)["encoding"]

            if not encoding:
                return None

            try:
                content = content.decode(encoding)
            except UnicodeDecodeError:
                # If the file cannot be decoded, return None
                return None

        # Extract the metadata
        line_count = content.count("\n") + 1
        size_in_bytes = file_info.file_size

        if category == "text":
            file_type = "text"
        else:
            file_type = "code"

        metadata = {
            # Use the file path as the unique ID
            "id": full_path,
            "file_extension": ext,
            "file_type": file_type,
            "category": category,
            "line_count": line_count,
            "size_in_bytes": size_in_bytes,
            "path": full_path,
        }
        return metadata, content

    def iterate(self, file_path: str):
        """
        Iterates over the files in a zip archive and yields the parsed content of each file.

        Args:
            file_path: The path to the zip archive.

        Yields:
            Parsed content of each file in the zip archive.
        """

        if not file_path:
            return

        with ZipFile(file_path, "r") as zip_ref:
            for file_info in zip_ref.infolist():
                filename = file_info.filename

                # Skip directories and hidden files
                if file_info.is_dir() or any(
                    part.startswith(".") for part in filename.split(os.sep)
                ):
                    continue

                parsed = self.parse_file(zip_ref, file_info)
                if parsed:
                    yield parsed


class GitHubExtractor(DocumentExtractor):
    def extract(self, content: str) -> dict:
        # Just return the content.
        return {"text": content}


class ArxivDownloader(DocumentDownloader):
    """
    A class for downloading article PDFs from arXiv.
    """

    def __init__(self, pdf_root_dir: str):
        """
        Initializes the DocBuilder object.

        Args:
            pdf_root_dir: The root directory for PDF repositories.
        """
        super().__init__()
        # The path under which the pdfs converted to text will be stored.
        self.pdf_root_dir = os.path.join(pdf_root_dir, "pdfs")
        os.makedirs(pdf_root_dir, exist_ok=True)
        os.makedirs(self.pdf_root_dir, exist_ok=True)

    def parse_id(self, input_string: str) -> str:
        """
        Extracts the arXiv ID from a given input string, which can be either an arXiv ID or a URL.

        Args:
            input_string: The input string that is either an arXiv ID or a URL.

        Returns:
            The extracted arXiv ID if the input is valid, otherwise None.
        """
        # Pattern to match a direct arXiv ID
        id_pattern = re.compile(r"\d{4}\.\d{4,5}(v\d+)?$")
        if id_pattern.match(input_string):
            return input_string

        # Pattern to match an arXiv URL and extract the ID
        url_pattern = re.compile(
            r"https?://(?:www\.)?arxiv\.org/(abs|pdf)/(\d{4}\.\d{4,5})(v\d+)?(\.pdf)?$"
        )
        url_match = url_pattern.match(input_string)
        if url_match:
            return url_match.group(2) + (
                url_match.group(3) if url_match.group(3) else ""
            )

        # Raise an error if the input does not match any of the expected formats
        raise ValueError(
            f"The provided input '{input_string}' does not match the expected URL or ID format."
        )

    def download(self, url: str) -> str:
        """
        Downloads the article as a PDF file.

        Args:
            url (str): The URL of the pdf.

        Returns:
            str: The path to the downloaded PDF file, or None if the download failed.
        """
        pdf_name = os.path.basename(url)
        pdf_file = os.path.join(self.pdf_root_dir, pdf_name)

        if os.path.exists(pdf_file):
            print(f"Article '{url}' already exists, skipping download.")
            return pdf_file

        article_id = self.parse_id(url)
        search_result = arxiv.Client().results(arxiv.Search(id_list=[article_id]))

        if article := next(search_result):
            print(f'Downloading arXiv article "{url}"...')
            pdf_path = article.download_pdf(
                dirpath=self.pdf_root_dir, filename=pdf_name
            )
        else:
            print(f"Failed to download article '{url}'.")
            return None

        return pdf_path


class ArxivIterator(DocumentIterator):
    """
    arXiv document iterator. Will go through the files and parse the supported ones.
    """

    def iterate(self, file_path: str):
        """
        Iterates over the pdf files and yields the parsed content of each file.

        Args:
            file_path: The path to the downloaded pdf.

        Yields:
            Parsed content of each file in the pdf.
        """
        if not file_path:
            return

        elements = partition(filename=file_path)

        # Extract the file name and extension in lower case.
        filename = os.path.basename(file_path)
        filename_no_ext, ext = os.path.splitext(filename)
        filename_no_ext = filename_no_ext.lower()
        ext = ext.lower()

        # Read and join the extracted content.
        content = "\n".join([str(el) for el in elements])

        # Extract the metadata
        line_count = content.count("\n") + 1
        size_in_bytes = os.path.getsize(file_path)
        metadata = {
            # Use the file path as the unique ID
            "id": file_path,
            "file_extension": ext,
            "file_type": "text",
            "category": "text",
            "line_count": line_count,
            "size_in_bytes": size_in_bytes,
            "path": file_path,
        }

        yield metadata, content


class ArxivExtractor(DocumentExtractor):
    def extract(self, content: str) -> dict:
        # Just return the content.
        return {"text": content}
